# -*- coding: utf-8 -*-

__version__ = "0.12.0"
__author__ = "Christopher Goes"
__email__ = "<goes8945@vandals.uidaho.edu>"
__license__ = "Apache 2.0"
__url__ = "https://github.com/GhostofGoes/ADLES"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Query information about a vSphere environment and objects within it.

Usage:
    vsphere-info [options]

Options:
    -h, --help          Prints this page
    --version           Prints current version
    -n, -no-color       Do not color terminal output
    -v, --verbose       Emit debugging logs to terminal
    -f, --file FILE     Name of JSON file with server connection information

Examples:
    vsphere-info -vf logins.json

"""

import logging

from docopt import docopt

from adles.vsphere import vm_utils, vsphere_utils
from adles.utils import script_setup, resolve_path, prompt_y_n_question
from adles.vsphere.folder_utils import enumerate_folder, format_structure

__version__ = "0.6.1"


def main():
    args = docopt(__doc__, version=__version__, help=True)
    server = script_setup('vsphere_info.log', args, (__file__, __version__))

    thing_type = str(input("What type of thing do you want to get information on?"
                           " (vm | datastore | vsphere | folder) "))

    # Single Virtual Machine
    if thing_type == "vm":
        vm, vm_name = resolve_path(server, "vm", "you want to get information on")
        logging.info(vm_utils.get_vm_info(vm, detailed=True, uuids=True, snapshot=True, vnics=True))

    # Datastore
    elif thing_type == "datastore":
        ds = server.get_datastore(str(input("Enter name of the Datastore [leave blank "
                                            "for first datastore found]: ")))
        logging.info(vsphere_utils.get_datastore_info(ds))

    # vCenter server
    elif thing_type == "vsphere":
        logging.info("%s", str(server))

    # Folder
    elif thing_type == "folder":
        folder, folder_name = resolve_path(server, "folder")
        if "VirtualMachine" in folder.childType \
                and prompt_y_n_question("Want to see power state of VMs in the folder?"):
            contents = enumerate_folder(folder, recursive=True, power_status=True)
        else:
            contents = enumerate_folder(folder, recursive=True, power_status=False)
        logging.info("Information for Folder %s\nTypes of items folder can contain: %s\n%s",
                     folder_name, str(folder.childType), format_structure(contents))

    # That's not a thing!
    else:
        logging.info("Invalid thing: %s", thing_type)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Power operations for Virtual Machines in vSphere.

Usage:
    vm-power [options]

Options:
    -h, --help          Prints this page
    --version           Prints current version
    -n, --no-color      Do not color terminal output
    -v, --verbose       Emit debugging logs to terminal
    -f, --file FILE     Name of JSON file with server connection information

Examples:
    vm-power -vf logins.json

"""

import logging

from docopt import docopt

from adles.utils import prompt_y_n_question, script_setup, resolve_path
from adles.vsphere.vm_utils import change_vm_state
from adles.vsphere.vsphere_utils import is_vm
from adles.vsphere.folder_utils import enumerate_folder, format_structure

__version__ = "0.3.6"


def main():
    args = docopt(__doc__, version=__version__, help=True)
    server = script_setup('vm_power.log', args, (__file__, __version__))

    operation = str(input("Enter the power operation you wish to perform"
                          " [on | off | reset | suspend]: "))
    attempt_guest = prompt_y_n_question("Attempt to use guest OS operations, if available? ")

    # TODO: prefixes
    # TODO: nesting
    if prompt_y_n_question("Multiple VMs? ", default="yes"):
        folder, folder_name = resolve_path(server, "folder", "with VMs")
        vms = [x for x in folder.childEntity if is_vm(x)]
        logging.info("Found %d VMs in folder '%s'", len(vms), folder_name)
        if prompt_y_n_question("Show the status of the VMs in the folder? "):
            logging.info("Folder structure: \n%s", format_structure(
                enumerate_folder(folder, recursive=True, power_status=True)))
        if prompt_y_n_question("Continue? "):
            for vm in vms:
                change_vm_state(vm, operation, attempt_guest)

    else:
        vm, vm_name = resolve_path(server, "VM")
        logging.info("Changing power state of '%s' to '%s'", vm_name, operation)
        change_vm_state(vm, operation, attempt_guest)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import docker  # TODO: add to requirements.txt once implemented

import adles.utils as utils


class DockerInterface:
    """ Generic interface for the Docker platform """

    __version__ = "0.1.2"

    def __init__(self, infra, spec):
        """
        
        :param infra: Dict of infrastructure information
        :param spec: Dict of a parsed specification
        """
        self._log = logging.getLogger('DockerInterface')
        self._log.debug("Initializing DockerInterface %s", DockerInterface.__version__)

        self.spec = spec
        self.metadata = spec["metadata"]
        self.services = spec["services"]
        self.networks = spec["networks"]
        self.folders = spec["folders"]
        self.infra = infra

        # TODO: this is initial testing of client
        # If needed, a wrapper class that simplifies the creation of containers will be made
        # Reference: https://docker-py.readthedocs.io/en/stable/client.html#client-reference
        # Initialize the Docker client
        self.client = docker.DockerClient(base_url=infra.get("url", "unix:///var/run/docker.sock"),
                                          tls=bool(infra.get("tls", True)))

        # Verify the connection to the client
        self.client.ping()

        # TODO: add interface-specific loggers so we know what interface is outputting
        self._log.debug("System info      : %s", str(self.client.info()))
        self._log.debug("System version   : %s", str(self.client.version()))

        # Authenticate to registry, if configured
        if "registry" in self.infra:
            reg = self.infra["registry"]
            reg_logins = utils.read_json(reg["login-file"])
            self.client.login(username=reg_logins["user"], password=reg_logins["pass"],
                              registry=reg["url"])

        # List images currently on the server
        self._log.debug("Images: %s", str(self.client.images.list()))

    def create_masters(self):
        pass

    def deploy_environment(self):
        pass

    def cleanup_masters(self, network_cleanup):
        pass

    def cleanup_environment(self, network_cleanup):
        pass



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from sys import exit
import os.path

import adles.vsphere.folder_utils as futils
import adles.vsphere.vm_utils as vm_utils
import adles.vsphere.vsphere_utils as vutils
from adles.utils import pad, read_json
from adles.vsphere import Vsphere
from adles.vsphere.network_utils import create_portgroup


class VsphereInterface:
    """ Generic interface for the VMware vSphere platform """

    __version__ = "0.8.6"

    # Names/prefixes
    master_prefix = "(MASTER) "
    master_root_name = "MASTER-FOLDERS"

    # Values at which to warn or error when exceeded
    # TODO: make these per-instance and configurable in spec?
    thresholds = {
        "folder": {
            "warn": 25,
            "error": 50},
        "service": {
            "warn": 50,
            "error": 70}
    }

    def __init__(self, infra, spec):
        """
        NOTE: it is assumed that the infrastructure and spec are both valid,
        and thus checks on key existence and types are not performed for REQUIRED elements.
        :param infra: Dict of infrastructure information
        :param spec: Dict of a parsed specification
        """
        self._log = logging.getLogger('VsphereInterface')
        self._log.debug("Initializing VsphereInterface %s", VsphereInterface.__version__)

        self.spec = spec
        self.metadata = spec["metadata"]
        self.services = spec["services"]
        self.networks = spec["networks"]
        self.folders = spec["folders"]
        self.infra = infra
        self.master_folder = None
        self.template_folder = None
        self.net_table = {}  # Used to do lookups of Generic networks during deployment

        # Read infrastructure login information
        if "login-file" in infra:
            logins = read_json(infra["login-file"])  # TODO: is this secure?
        else:
            self._log.warning("No login-file specified, defaulting to user prompts...")
            logins = {}

        # Instantiate the vSphere vCenter server instance class
        self.server = Vsphere(username=logins.get("user"),
                              password=logins.get("pass"),
                              hostname=infra.get("hostname"),
                              port=int(infra.get("port")),
                              datastore=infra.get("datastore"),
                              datacenter=infra.get("datacenter"))

        # Acquire ESXi hosts (TODO: make these their own classes /w network, etc methods)
        if "hosts" in infra:
            hosts = infra["hosts"]
            self.host = self.server.get_host(hosts[0])  # TODO: temporary hack
            self.hosts = [self.server.get_host(h) for h in hosts]  # Gather all the ESXi hosts
        else:
            self.host = self.server.get_host()  # First host found in Datacenter

        # Instantiate and initialize Groups
        self.groups = self._init_groups()

        # Set the server root folder (TODO: network folder in infrastructure spec)
        if "server-root" in infra:
            self.server_root = self.server.get_folder(infra["server-root"])
            if not self.server_root:
                self._log.error("Could not find server-root folder '%s'", infra["server-root"])
                exit(1)
        else:
            self.server_root = self.server.datacenter.vmFolder  # Default to Datacenter VM folder
        self._log.info("Server root folder: %s", self.server_root.name)

        # Set environment root folder (TODO: this can be simplified and consolidated)
        if "folder-name" not in self.metadata:
            self.root_path, self.root_name = ("", self.metadata["name"])
            self.root_folder = futils.traverse_path(folder=self.server_root,
                                                    path=self.root_name,
                                                    generate=True)
        else:
            self.root_path, self.root_name = os.path.split(self.metadata["folder-name"])
            self.root_folder = futils.traverse_path(folder=self.server_root,
                                                    path=self.metadata["folder-name"],
                                                    generate=True)

        self._log.debug("Environment root folder name: %s", self.root_name)
        if not self.root_folder:  # Create if it's not found
            parent = futils.traverse_path(self.server_root, self.root_path)
            self.root_folder = self.server.create_folder(self.root_name, parent)
            if not self.root_folder:
                self._log.error("Could not create root folder '%s'", self.root_name)
                exit(1)
        self._log.info("Environment root folder: %s", self.root_folder.name)

        # Set default vSwitch name
        if "vswitch" in infra:
            self.vswitch_name = infra["vswitch"]
        else:  # TODO: is this a good default? (If you have to ask this, it's probably not)
            from pyVmomi import vim
            self.vswitch_name = self.server.get_item(vim.Network).name

        self._log.debug("Finished initializing VsphereInterface")

    def _init_groups(self):
        """
        Instantiate and initialize Groups
        :return: Dict of Groups
        """
        from adles.group import Group, get_ad_groups
        groups = {}

        # Instantiate Groups
        for name, config in self.spec["groups"].items():
            if "instances" in config:  # Template groups
                groups[name] = [Group(name, config, i) for i in range(1, config["instances"] + 1)]
            else:  # Standard groups
                groups[name] = Group(name=name, group=config)

        # Initialize Active Directory-type Group user names
        ad_groups = get_ad_groups(groups)
        for g in ad_groups:
            # res = self.server.get_users(belong_to_group=g.ad_group, find_users=True) (TODO)
            res = None
            if res is not None:
                for r in res:
                    if r.group is True:  # Reference: pyvmomi/docs/vim/UserSearchResult.rst
                        self._log.error("Result '%s' is not a user", str(r))
                    else:
                        g.users.append(r.principal)
                g.size = (len(g.users) if len(g.users) > 1 else 1)  # Set the size, default to 1
            else:
                self._log.error("Could not initialize AD-group %s", str(g.ad_group))

        if hasattr(self.server.user_dir, "domainList"):
            self._log.debug("Domains on server: %s", str(self.server.user_dir.domainList))

        return groups

    def create_masters(self):
        """ Exercise Environment Master creation phase """

        # Get folder containing templates
        self.template_folder = futils.traverse_path(self.server_root,
                                                    self.infra["template-folder"])
        if not self.template_folder:
            self._log.error("Could not find template folder in path '%s'",
                            self.infra["template-folder"])
            return
        else:
            self._log.debug("Found template folder: '%s'", self.template_folder.name)

        # Create master folder to hold base service instances
        self.master_folder = self.server.create_folder(
            VsphereInterface.master_root_name, self.root_folder)
        self._log.info("Created master folder '%s' under folder '%s'",
                       VsphereInterface.master_root_name, self.root_name)

        # TODO: implement configuration of "network-interface" in the "services" top-level section
        # Create networks for master instances
        for net in self.networks:  # Iterate through the base types
            self._create_master_networks(net_type=net, default_create=True)

        # Create Master instances
        # TODO: Apply master-group permissions [default: group permissions]
        self._master_parent_folder_gen(self.folders, self.master_folder)

        # Output fully deployed master folder tree to debugging
        self._log.debug(futils.format_structure(
            futils.enumerate_folder(self.root_folder)))

    def _master_parent_folder_gen(self, folder, parent):
        """
        Generates parent-type Master folders
        :param folder: Dict with the folder tree structure as in spec
        :param parent: Parent vim.Folder
        """

        group = None
        master_group = None

        # We have to check every item, as they could be either keywords or sub-folders
        for sub_name, sub_value in folder.items():
            if sub_name == "instances":
                pass  # Instances don't matter for the Master phase
            elif sub_name == "description":
                pass  # Note: could use Tags or Custom Attributes to add descriptions to objects
            elif sub_name == "group":
                group = self._get_group(sub_value)
            elif sub_name == "master-group":
                master_group = self._get_group(sub_value)
            else:
                folder_name = VsphereInterface.master_prefix + sub_name
                new_folder = self.server.create_folder(folder_name, create_in=parent)

                if "services" in sub_value:  # It's a base folder
                    self._log.info("Generating Master base-type folder %s", sub_name)
                    self._master_base_folder_gen(sub_name, sub_value, new_folder)
                else:  # It's a parent folder, recurse
                    self._log.info("Generating Master parent-type folder %s", sub_name)
                    self._master_parent_folder_gen(sub_value, parent=new_folder)

        # TODO: apply master group permissions
        if master_group is None:
            # Since group is required, we can safely assume it's been set (TODO: NO ITS NOT!)
            master_group = group

    def _master_base_folder_gen(self, folder_name, folder_dict, parent):
        """
        Generates base-type Master folders
        :param folder_name: Name of the base folder
        :param folder_dict: Dict with the base folder tree as in spec
        :param parent: Parent vim.Folder
        """
        # Set the group to apply permissions for (TODO: apply permissions)
        if "master-group" in folder_dict:
            master_group = self._get_group(folder_dict["master-group"])
        else:
            master_group = self._get_group(folder_dict["group"])

        # Create Master instances
        for sname, sconfig in folder_dict["services"].items():
            if not self._is_vsphere(sconfig["service"]):
                self._log.debug("Skipping non-vsphere service '%s'", sname)
                continue

            self._log.info("Creating Master instance '%s' from service '%s'",
                           sname, sconfig["service"])

            vm = self._clone_service(parent, sconfig["service"])

            if not vm:
                self._log.error("Failed to create Master instance '%s' in folder '%s'",
                                sname, folder_name)
                continue  # Skip to the next service

            # NOTE: management interfaces matter here!
            self._configure_nics(vm, networks=sconfig["networks"])  # Configure VM NICs

            # Post-creation snapshot
            vm_utils.create_snapshot(vm, "initial mastering snapshot",
                                     "Beginning of Master configuration")

    def _clone_service(self, folder, service_name):
        """
        Retrieves and clones a service into a master folder
        :param folder: vim.Folder to clone into
        :param service_name: Name of the service to clone
        :return: The service vim.VirtualMachine instance
        """
        if not self._is_vsphere(service_name):
            self._log.debug("Skipping non-vsphere service '%s'", service_name)
            return None

        config = self.services[service_name]

        self._log.debug("Cloning service '%s'", service_name)
        vm_name = VsphereInterface.master_prefix + service_name

        # Find the template that matches the service definition
        template = futils.traverse_path(self.template_folder, config["template"])
        if not template:
            self._log.error("Could not find template '%s' for service '%s'",
                            config["template"], service_name)
            return None

        # Clone the template to create the Master instance
        vm_utils.clone_vm(template, folder=folder, name=vm_name,
                          clone_spec=self.server.gen_clone_spec())

        # Get new cloned instance
        vm = futils.traverse_path(folder, vm_name)
        if vm:
            self._log.debug("Successfully cloned service '%s' to folder '%s'",
                            service_name, folder.name)
            if "note" in config:  # Set VM note if specified
                vm_utils.set_note(vm, note=config["note"])
            return vm
        else:
            self._log.error("Failed to clone VM '%s' for service '%s'", vm_name, service_name)
            return None

    def _create_master_networks(self, net_type, default_create):
        """
        Creates a network as part of the Master creation phase
        :param net_type: Top-level type of the network (unique | generic | base)
        :param default_create: Whether to create networks if they don't already exist
        """
        self.host.configManager.networkSystem.RefreshNetworkSystem()  # Pick up any recent changes
        self._log.info("Creating %s", net_type)

        for name, config in self.networks[net_type].items():
            exists = self.server.get_network(name)
            if exists:
                self._log.debug("PortGroup '%s' already exists on host '%s'", name, self.host.name)
            else:  # NOTE: if monitoring, we want promiscuous=True
                self._log.warning("PortGroup '%s' does not exist on host '%s'",
                                  name, self.host.name)
                if default_create:
                    self._log.debug("Creating portgroup '%s' on host '%s'", name, self.host.name)
                    create_portgroup(name=name, host=self.host, promiscuous=False,
                                     vlan=int(config.get("vlan", next(self._get_vlan()))),
                                     vswitch_name=config.get("vswitch", self.vswitch_name))

    def _configure_nics(self, vm, networks, instance=None):
        """
        Configures Network Interfaces for a service instance
        :param vm: vim.VirtualMachine
        :param networks: List of networks to configure
        :param instance: Current instance of a folder for Deployment purposes
        """
        self._log.debug("Editing NICs for VM '%s'", vm.name)
        num_nics = len(list(vm.network))
        num_nets = len(networks)
        nets = networks  # Copy the passed variable so we can edit it later

        # Ensure number of NICs on VM matches number of networks configured for the service
        # Note that monitoring interfaces will be counted and included in the networks list
        if num_nics > num_nets:     # Remove excess interfaces
            diff = int(num_nics - num_nets)
            self._log.debug("VM '%s' has %d extra NICs, removing...", vm.name, diff)
            for i, nic in zip(range(1, diff + 1), reversed(range(num_nics))):
                vm_utils.delete_nic(vm, nic_number=nic)

        elif num_nics < num_nets:   # Create missing interfaces
            diff = int(num_nets - num_nics)
            self._log.debug("VM '%s' is deficient %d NICs, adding...", vm.name, diff)
            for i in range(diff):   # Add NICs to VM and pop them from the list of networks
                nic_model = ("vmxnet3" if vm_utils.has_tools(vm) else "e1000")
                net_name = nets.pop()
                vm_utils.add_nic(vm, network=self.server.get_network(net_name),
                                 model=nic_model, summary=net_name)
            num_nets = len(networks)

        # Edit the interfaces
        # NOTE: any NICs that were added earlier shouldn't be affected by this
        # TODO: traverse folder to get network? (need to switch to DVswitches I think)
        for net_name, i in zip(networks, range(1, num_nets + 1)):
            # Setting the summary to network name allows viewing of name without requiring
            # read permissions to the network itself
            if instance is not None:  # Resolve generic networks for deployment phase
                net_name = self._get_net(net_name, instance)
            network = self.server.get_network(net_name)
            if vm_utils.get_nic_by_id(vm, i).backing.network == network:
                continue  # Skip NICs that are already configured
            else:
                vm_utils.edit_nic(vm, nic_id=i, port_group=network, summary=net_name)

    def deploy_environment(self):
        """ Exercise Environment deployment phase """
        # Get the master folder root
        self.master_folder = futils.traverse_path(self.root_folder,
                                                  VsphereInterface.master_root_name)
        if self.master_folder is None:
            self._log.error("Could not find Master folder '%s'. "
                            "Please ensure the  Master Creation phase has been run "
                            "and the folder exists before attempting Deployment",
                            VsphereInterface.master_root_name)
            exit(1)
        self._log.debug("Master folder name: %s\tPrefix: %s",
                        self.master_folder.name, VsphereInterface.master_prefix)

        # Verify and convert Master instances to templates
        self._log.info("Converting Masters to Templates")
        self._convert_and_verify(folder=self.master_folder)
        self._log.info("Finished converting Masters to Templates")

        self._log.info("Deploying environment...")
        self._deploy_parent_folder_gen(spec=self.folders, parent=self.root_folder, path="")
        self._log.info("Finished deploying environment")

        # Output fully deployed environment tree to debugging
        self._log.debug(futils.format_structure(futils.enumerate_folder(self.root_folder)))

    def _convert_and_verify(self, folder):
        """
        Converts masters to templates before deployment.
        This also ensures they are powered off before being cloned.
        :param folder: vim.Folder
        """
        self._log.debug("Converting Masters in folder '%s' to templates", folder.name)
        for item in folder.childEntity:
            if vutils.is_vm(item):
                if vm_utils.is_template(item):  # Skip if they've already been created
                    self._log.debug("Master '%s' is already a template", item.name)
                    continue

                # Cleanly power off VM before converting to template
                if vm_utils.powered_on(item):
                    vm_utils.change_vm_state(item, "off", attempt_guest=True)

                # Take a snapshot to allow reverts to start of exercise
                vm_utils.create_snapshot(item, "Start of exercise",
                                         "Beginning of deployment phase, post-master configuration")

                # Convert master to template
                vm_utils.convert_to_template(item)
                self._log.debug("Converted Master '%s' to Template. Verifying...", item.name)

                # Check if it successfully converted to a snapshot
                if not vm_utils.is_template(item):
                    self._log.error("Master '%s' did not convert to template", item.name)
                else:
                    self._log.debug("Verified!")
            elif vutils.is_folder(item):  # Recurse into sub-folders
                self._convert_and_verify(item)
            else:
                self._log.debug("Unknown item found while converting Masters to templates: %s",
                                str(item))

    def _deploy_parent_folder_gen(self, spec, parent, path):
        """
        Generates parent-type folder trees
        :param spec: Dict with folder specification
        :param parent: Parent vim.Folder
        :param path: Folders path at the current level
        """
        group = None  # TODO: apply group permissions

        for sub_name, sub_value in spec.items():
            if sub_name == "instances":
                pass  # The instances have already been handled
            elif sub_name == "description":
                pass  # NOTE: may use Tags or Custom Attributes to add descriptions to objects
            elif sub_name == "group":
                group = self._get_group(sub_value)
            elif sub_name == "master-group":
                pass  # Not applicable for Deployment phase
            else:
                num_instances, prefix = self._instances_handler(spec=spec, obj_name=sub_name,
                                                                obj_type="folder")

                # Create instances of the parent folder
                self._log.debug("Deploying parent-type folder '%s'", sub_name)
                for i in range(num_instances):
                    # If prefix is undefined or there's a single instance, use the folder's name
                    instance_name = (sub_name if prefix == "" or num_instances == 1 else prefix)

                    # If multiple instances, append padded instance number
                    instance_name += (pad(i) if num_instances > 1 else "")

                    # Create a folder for the instance
                    new_folder = self.server.create_folder(instance_name, create_in=parent)

                    if "services" in sub_value:  # It's a base folder
                        self._deploy_base_folder_gen(folder_name=sub_name, folder_items=sub_value,
                                                     parent=new_folder,
                                                     path=self._path(path, sub_name))
                    else:  # It's a parent folder
                        self._deploy_parent_folder_gen(parent=new_folder, spec=sub_value,
                                                       path=self._path(path, sub_name))

    def _deploy_base_folder_gen(self, folder_name, folder_items, parent, path):
        """
        Generates folder tree for deployment stage
        :param folder_name: Name of the folder
        :param folder_items: Dict of items in the folder
        :param parent: Parent vim.Folder
        :param path: Folders path at the current level
        """
        # Set the group to apply permissions for (TODO: apply permissions)
        group = self._get_group(folder_items["group"])

        # Get number of instances and check if it exceeds configured limits
        num_instances, prefix = self._instances_handler(spec=folder_items, obj_name=folder_name,
                                                        obj_type="folder")

        # Create instances
        self._log.info("Deploying base-type folder '%s'", folder_name)
        for i in range(num_instances):
            # If no prefix is defined or there's only a single instance, use the folder's name
            instance_name = (folder_name if prefix == "" or num_instances == 1 else prefix)

            # If multiple instances, append padded instance number
            instance_name += (pad(i) if num_instances > 1 else "")

            if num_instances > 1:  # Create a folder for the instance
                new_folder = self.server.create_folder(instance_name, create_in=parent)
            else:  # Don't duplicate folder name for single instances
                new_folder = parent

            # Use the folder's name for the path, as that's what matches the Master version
            self._log.info("Generating services for base-type folder instance '%s'", instance_name)
            self._deploy_gen_services(services=folder_items["services"], parent=new_folder,
                                      path=self._path(path, folder_name), instance=i)

    def _deploy_gen_services(self, services, parent, path, instance):
        """
        Generates the services in a folder
        :param services: The "services" dict in a folder
        :param parent: Parent vim.Folder
        :param path: Folders path at the current level
        :param instance: What instance of a base folder this is
        """
        # Iterate through the services
        for service_name, value in services.items():
            # Ignore non-vsphere services
            if not self._is_vsphere(value["service"]):
                self._log.debug("Skipping non-vsphere service '%s'", service_name)
                continue
            self._log.info("Generating service '%s' in folder '%s'", service_name, parent.name)

            # Get number of instances for the service and check if it exceeds configured limits
            num_instances, prefix = self._instances_handler(spec=value, obj_name=service_name,
                                                            obj_type="service")

            # Get the Master template instance to clone from
            service = futils.traverse_path(self.master_folder, self._path(path, value["service"]))
            if service is None:  # Check if the lookup was successful
                self._log.error("Could not find Master instance for service '%s' in this path:\n%s",
                                value["service"], path)
                continue  # Skip to the next service

            # Clone the instances of the service from the master
            for i in range(num_instances):
                instance_name = prefix + service_name + (" " + pad(i) if num_instances > 1 else "")
                vm_utils.clone_vm(vm=service, folder=parent, name=instance_name,
                                  clone_spec=self.server.gen_clone_spec())
                vm = futils.traverse_path(parent, instance_name)
                if vm:
                    self._configure_nics(vm=vm, networks=value["networks"], instance=instance)
                else:
                    self._log.error("Could not find cloned instance '%s' in folder '%s'",
                                    instance_name, service_name, parent.name)

    @staticmethod
    def _path(path, name):
        """
        Generates next step of the path for deployment of Masters
        :param path: Current path
        :param name: Name to add to the path
        :return: The updated path
        """
        return str(path + '/' + VsphereInterface.master_prefix + name)

    def _instances_handler(self, spec, obj_name, obj_type):
        """
        Determines number of instances and optional prefix using specification
        :param spec: Dict of folder
        :param obj_name: Name of the thing being handled
        :param obj_type: Type of the thing being handled (folder | service)
        :return: (Number of instances, Prefix)
        """
        # TODO: move this into base Interface class
        num = 1
        prefix = ""
        if "instances" in spec:
            if type(spec["instances"]) == int:
                num = int(spec["instances"])
            else:
                if "prefix" in spec["instances"]:
                    prefix = str(spec["instances"]["prefix"])

                if "number" in spec["instances"]:
                    num = int(spec["instances"]["number"])
                elif "size-of" in spec["instances"]:
                    # size_of = spec["instances"]["size-of"]
                    # num = int(self._get_group(size_of).size)
                    # if num < 1:
                    num = 1  # TODO: WORKAROUND FOR AD-GROUPS
                else:
                    self._log.error("Unknown instances specification: %s", str(spec["instances"]))
                    num = 0

        # Check if the number of instances exceeds the configured thresholds for the interface
        if num > self.thresholds[obj_type]["error"]:
            self._log.error("%d instances of %s '%s' is beyond the configured %s threshold of %d",
                            num, obj_type, obj_name, self.__name__,
                            self.thresholds[obj_type]["error"])
            raise Exception("Threshold exception")
        elif num > self.thresholds[obj_type]["warn"]:
            self._log.warning("%d instances of %s '%s' is beyond the configured %s threshold of %d",
                              num, obj_type, obj_name, self.__name__,
                              self.thresholds[obj_type]["warn"])

        return num, prefix

    def _get_group(self, group_name):
        """
        Provides a uniform way to get information about normal groups and template groups
        :param group_name: Name of the group
        :return: Group object
        """
        from adles.group import Group
        if group_name in self.groups:
            g = self.groups[group_name]
            if isinstance(g, Group):    # Normal groups
                return g
            elif isinstance(g, list):   # Template groups
                return g[0]
            else:
                self._log.error("Unknown type for group '%s': %s", str(group_name), str(type(g)))
        else:
            self._log.error("Could not get group '%s' from VsphereInterface groups", group_name)

    def _is_vsphere(self, service_name):
        """
        Checks if a service instance is defined as a vSphere service
        :param service_name: Name of the service to lookup in list of defined services
        :return: bool
        """
        # TODO: make "template" and other platform identifiers global keywords
        if service_name not in self.services:
            self._log.error("Could not find service %s in the list of defined services",
                            service_name)
        elif "template" in self.services[service_name]:
            return True
        return False

    def _determine_net_type(self, network_label):
        """
        Determines the type of a network
        :param network_label: Name of the network
        :return: Type of the network ("generic-networks" | "unique-networks")
        """
        for net_name, net_value in self.networks.items():
            # TODO: when checking syntax, need to check networks match/exist with case sensitivity
            vals = set(k.lower() for k in net_value)  # Case-insensitive comparisons
            if network_label.lower() in vals:
                return net_name
        self._log.error("Couldn't find type for network '%s'", network_label)
        return ""

    @staticmethod
    def _get_vlan():
        """
        Generates unique VLAN tags
        :return: int VLAN tag 
        """
        for i in range(2000, 4096):
            yield i

    def _get_net(self, name, instance=-1):
        """
        Resolves network names. This is mainly to handle generic-type networks.
        If a generic network does not exist, it is created and added to the interface lookup table.
        :param name: Name of the network
        :param instance: Instance number (Only applies to generic-type networks)
        :return: Resolved network name
        """
        # TODO: could use this to do network lookups on the server as well
        net_type = self._determine_net_type(name)
        if net_type == "unique-networks":
            return name
        elif net_type == "generic-networks":
            if instance == -1:
                self._log.error("Invalid instance for _get_net: %d", instance)
                raise ValueError
            net_name = name + "-GENERIC-" + pad(instance)  # Generate full name for the generic net
            if net_name not in self.net_table:
                exists = self.server.get_network(net_name)
                if exists is not None:
                    self._log.debug("PortGroup '%s' already exists on host '%s'", net_name,
                                    self.host.name)
                else:  # Create the generic network if it does not exist
                    self._log.debug("Creating portgroup '%s' on host '%s'", net_name,
                                    self.host.name)
                    vsw = self.networks["generic-networks"][name].get("vswitch", self.vswitch_name)
                    create_portgroup(name=net_name, host=self.host, promiscuous=False,
                                     vlan=next(self._get_vlan()), vswitch_name=vsw)
                self.net_table[net_name] = True  # Register the existence of the generic
            return net_name
        else:
            self._log.error("Invalid network type %s for network %s", net_type, name)
            raise TypeError

    def cleanup_masters(self, network_cleanup=False):
        """ Cleans up any master instances"""

        # TODO: look at getorphanedvms in pyvmomi-community-samples for how to do this
        # Get the folder to cleanup in
        master_folder = futils.find_in_folder(self.root_folder, self.master_root_name)
        self._log.info("Found master folder '%s' under folder '%s', proceeding with cleanup...",
                       master_folder.name, self.root_folder.name)

        # Recursively descend from master folder, destroying anything with the prefix
        futils.cleanup(folder=master_folder, vm_prefix=self.master_prefix,
                       recursive=True, destroy_folders=True, destroy_self=True)

        # Cleanup networks (TODO: use network folders to aid in this, during creation phase)
        if network_cleanup:
            pass

    def cleanup_environment(self, network_cleanup=False):
        """ Cleans up a deployed environment """

        # Get the root environment folder to cleanup in
        enviro_folder = self.root_folder

        # TODO: ensure master folder is skipped

        # Cleanup networks (TODO: use network folders to aid in this, during creation phase)
        if network_cleanup:
            pass

    def __repr__(self):
        return str("VsphereInterface(%s,%s)" % (str(self.infra), str(self.spec)))

    def __str__(self):
        return str(self.server) + str(self.groups) + str(self.hosts)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.server == other.server and \
               self.groups == other.groups and self.hosts == other.hosts

    def __ne__(self, other):
        return not self.__eq__(other)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from pyVmomi import vim

from adles.utils import check, split_path
from adles.vsphere.vsphere_utils import is_folder, is_vm, wait_for_task


@check(vim.Folder, "folder")
def get_in_folder(folder, name, recursive=False, vimtype=None):
    """
    Retrieves an item from a datacenter folder
    :param folder: vim.Folder to search in
    :param name: Name of object to find
    :param recursive: Recurse into sub-folders [default: False]
    :param vimtype: Type of object to search for [default: None]
    :return: Object found or None if nothing was found
    """
    item = None
    if name is not None:
        item = find_in_folder(folder, name, recursive=recursive, vimtype=vimtype)
    if item is None:  # Get first item found of type if can't find in folder or name isn't specified
        if len(folder.childEntity) > 0 and vimtype is None:
            return folder.childEntity[0]
        elif len(folder.childEntity) > 0:
            for i in folder.childEntity:
                if isinstance(i, vimtype):
                    return i
            logging.error("Could not find item of type '%s' in folder '%s'",
                          vimtype.__name__, folder.name)
        else:
            logging.error("There are no items in folder %s", folder.name)
    return item


@check(vim.Folder, "folder")
def find_in_folder(folder, name, recursive=False, vimtype=None):
    """
    Finds and returns an specific object in a folder
    :param folder: vim.Folder object to search in
    :param name: Name of the object to find
    :param recursive: Recurse into sub-folders [default: False]
    :param vimtype: Type of object to search for [default: None]
    :return: Object found or None if nothing was found
    """
    item_name = name.lower()  # Convert to lowercase for case-insensitive comparisons
    found = None
    for item in folder.childEntity:
        if hasattr(item, 'name') and item.name.lower() == item_name:  # Check if the name matches
            if vimtype is not None and not isinstance(item, vimtype):
                continue
            found = item
        elif recursive and is_folder(item):  # Recurse into sub-folders
            found = find_in_folder(item, name=item_name, recursive=recursive, vimtype=vimtype)
        if found is not None:
            return found
    return None


@check(vim.Folder, "folder")
def traverse_path(folder, path, lookup_root=None, generate=False):
    """
    Traverses a folder path to find a object with a specific name
    :param folder: vim.Folder to search in
    :param path: String with path in POSIX format (Templates/Windows/ to get the 'Windows' folder)
    :param lookup_root: If root of path is not found in folder, lookup using this Vsphere object
    :param generate: Parts of the path that do not exist are created.
    :return: Object at end of path
    """
    logging.debug("Traversing path '%s' from folder '%s'", path, folder.name)
    folder_path, name = split_path(path)

    # Check if root of the path is in the folder
    # This is to allow relative paths to be used if lookup_root is defined
    folder_items = [x.name.lower() for x in folder.childEntity if hasattr(x, 'name')]
    if len(folder_path) > 0 and folder_path[0] not in folder_items:
        if lookup_root is not None:
            logging.debug("Root %s not in folder %s, looking up...",
                          folder_path[0], folder.name)
            folder = lookup_root.get_folder(folder_path.pop(0))  # Lookup the path root on server
        else:
            logging.error("Could not find root '%s' of path '%s' in folder '%s'",
                          folder_path[0], path, folder.name)
            return None

    current = folder  # Start with the defined folder
    for f in folder_path:  # Try each folder name in the path
        found = None
        for item in current.childEntity:  # Iterate through items in the current folder
            if is_folder(item) and item.name.lower() == f:  # If Folder is part of path
                found = item  # This is the next folder in the path
                break  # Break to outer loop to check this folder for the next part of the path
        if generate and found is None:  # Can't find the folder, so create it
            logging.warning("Generating folder %s in path", f)
            create_folder(folder, f)  # Generate the folder
        elif found is not None:
            current = found

    if name != '':  # Since the split had a basename, look for an item with matching name
        return find_in_folder(current, name)
    else:  # No basename, so just return whatever was at the end of the path
        return current


@check(vim.Folder, "folder")
def enumerate_folder(folder, recursive=True, power_status=False):
    """
    Enumerates a folder structure and returns the result as a python object with the same structure
    :param folder: vim.Folder
    :param recursive: Whether to recurse into any sub-folders [default: True]
    :param power_status: Display
    :return: The nested python object with the enumerated folder structure
    """
    children = []
    for item in folder.childEntity:
        if is_folder(item):
            if recursive:  # Recurse into sub-folders and append the resultant sub-tree
                children.append(enumerate_folder(item, recursive))
            else:  # Don't recurse, just append the folder
                children.append('- ' + item.name)
        elif is_vm(item):
            if power_status:
                if item.runtime.powerState == vim.VirtualMachine.PowerState.poweredOn:
                    children.append('* ON  ' + item.name)
                elif item.runtime.powerState == vim.VirtualMachine.PowerState.poweredOff:
                    children.append('* OFF ' + item.name)
                elif item.runtime.powerState == vim.VirtualMachine.PowerState.suspended:
                    children.append('* SUS ' + item.name)
                else:
                    logging.error("Invalid power state for VM: %s", item.name)
            else:
                children.append('* ' + item.name)
        else:
            children.append("UNKNOWN ITEM: %s" % str(item))
    return '+ ' + folder.name, children  # Return tuple of parent and children


# Similar to: https://docs.python.org/3/library/pprint.html
def format_structure(structure, indent=4, _depth=0):
    """
    Converts a nested structure of folders into a formatted string
    :param structure: structure to format
    :param indent: Number of spaces to indent each level of nesting [default: 4]
    :param _depth: Current depth (USED INTERNALLY BY FUNCTION)
    :return: Formatted string
    """
    fmat = ""
    newline = '\n' + str(_depth * str(indent * ' '))

    if type(structure) == tuple:
        fmat += newline + str(structure[0])
        fmat += format_structure(structure[1], indent, _depth + 1)
    elif type(structure) == list:
        for item in structure:
            fmat += format_structure(item, indent, _depth)
    elif type(structure) == str:
        fmat += newline + structure
    else:
        logging.error("Unexpected type in folder structure for item '%s': %s",
                      str(structure), type(structure))
    return fmat


@check(vim.Folder, "folder")
def create_folder(folder, folder_name):
    """
    Creates a VM folder in the specified folder
    :param folder: vim.Folder object to create the folder in
    :param folder_name: Name of folder to create
    :return: The created vim.Folder object
    """
    exists = find_in_folder(folder, folder_name)  # Check if the folder already exists
    if exists:
        logging.warning("Folder '%s' already exists in folder '%s'", folder_name, folder.name)
        return exists  # Return the folder that already existed
    else:
        logging.debug("Creating folder '%s' in folder '%s'", folder_name, folder.name)
        try:
            return folder.CreateFolder(folder_name)  # Create the folder and return it
        except vim.fault.DuplicateName as dupe:
            logging.error("Could not create folder '%s' in '%s': folder already exists as '%s'",
                          folder_name, folder.name, dupe.name)
        except vim.fault.InvalidName as invalid:
            logging.error("Could not create folder '%s' in '%s': Invalid folder name '%s'",
                          folder_name, folder.name, invalid.name)
    return None


@check(vim.Folder, "folder")
def cleanup(folder, vm_prefix='', folder_prefix='', recursive=False,
            destroy_folders=False, destroy_self=False):
    """
    Destroys VMs and folders under the specified folder.
    :param folder: vim.Folder object to clean
    :param vm_prefix: Only destroy VMs with names starting with the prefix [default: '']
    :param folder_prefix: Only destroy or search in folders with names starting with the prefix
    [default: '']
    :param recursive: Recursively descend into any sub-folders [default: False]
    :param destroy_folders: Destroy folders in addition to VMs [default: False]
    :param destroy_self: Destroy the folder specified [default: False]
    """
    logging.debug("Cleaning folder '%s'", folder.name)
    import adles.vsphere.vm_utils as vm_utils  # Prevents module-level import issues

    for item in folder.childEntity:
        if is_vm(item) and str(item.name).startswith(vm_prefix):  # Handle VMs
            vm_utils.destroy_vm(vm=item)  # Delete the VM from the Datastore
        elif is_folder(item) and str(item.name).startswith(folder_prefix):   # Handle folders
            if destroy_folders:  # Destroys folder and ALL of it's sub-objects
                cleanup(item, destroy_folders=True, destroy_self=True)
            elif recursive:  # Simply recurses to find more items
                cleanup(item, vm_prefix=vm_prefix, folder_prefix=folder_prefix, recursive=True)

    # Note: UnregisterAndDestroy does NOT delete VM files off the datastore
    # Only use if folder is already empty!
    if destroy_self:
        logging.debug("Destroying folder: '%s'", folder.name)
        wait_for_task(folder.UnregisterAndDestroy_Task())


@check(vim.Folder, "folder")
def retrieve_items(folder, vm_prefix='', folder_prefix='', recursive=False):
    """
    Retrieves VMs and folders from a folder structure
    :param folder: vim.Folder to begin search in (Note: it is NOT returned in list of folders)
    :param vm_prefix: VM prefix to search for [default: None]
    :param folder_prefix: Folder prefix to search for [default: None]
    :param recursive: Recursively descend into sub-folders
           (Note: This will recurse regardless of folder prefix!) [default: False]
    :return: ([vms], [folders])
    """
    vms = []
    folders = []

    for item in folder.childEntity:  # Iterate through all items in the folder
        if is_vm(item) and str(item.name).startswith(vm_prefix):
            vms.append(item)  # Add matching vm to the list
        elif is_folder(item):
            if str(item.name).startswith(folder_prefix):
                folders.append(item)  # Add matching folder to the list
            if recursive:  # Recurse into sub-folders
                v, f = retrieve_items(item, vm_prefix, folder_prefix, recursive)
                vms.extend(v)
                folders.extend(f)
    return vms, folders



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging


class Group:
    """ Manages a group of users that has been loaded from a specification """

    def __init__(self, name, group, instance=None):
        """
        :param name: Name of the group
        :param group: Dict specification of the group
        :param instance: Instance number of a template group [default: None]
        """
        self._log = logging.getLogger('Group')
        self._log.debug("Initializing Group '%s'", name)

        if type(group) != dict:
            self._log.error("Class Group must be initialized with a dict, not a %s. "
                            "\nThe offending object: %s", type(group).__name__, str(group))
            raise Exception()

        if instance:
            self.is_template = True
            self.instance = instance
        else:
            self.is_template = False

        # !!! NOTE: ad-groups must be handled externally by caller !!!
        if "ad-group" in group:
            group_type = "ad"
            self.ad_group = group["ad-group"]
            users = []
            if instance:  # Template groups
                self.ad_group += " " + str(instance)  # This is the " X" in the spec

        elif "filename" in group:
            from adles.utils import read_json
            group_type = "standard"
            if instance:    # Template group
                users = [(user, pw)
                         for user, pw in read_json(group["filename"])[str(instance)].items()]
            else:           # Standard group
                users = [(user, pw)
                         for user, pw in read_json(group["filename"]).items()]

        elif "user-list" in group:
            group_type = "standard"
            users = group["user-list"]

        else:
            self._log.error("Invalid group dict for group '%s': %s", name, str(group))
            raise Exception()

        self.group_type = group_type
        self.users = users
        self.size = int(len(self.users))
        self.name = str(name)
        self._log.debug("Finished initializing Group '%s' with %d users",
                        self.name, self.size)


def get_ad_groups(groups):
    """
    Extracts Active Directory-type groups from a dict of groups
    :param groups: Dict of groups and lists of groups
    :return: List of AD groups extracted
    """
    ad_groups = []
    for _, g in groups.items():  # Ignore the group name, nab the group
        if type(g) == list:
            for i in g:
                if type(i) == Group:
                    if i.group_type == "ad":
                        ad_groups.append(i)
        elif type(g) == Group:
            if g.group_type == "ad":
                ad_groups.append(g)
        else:
            logging.error("Invalid type '%s' for a group in get_ad_groups: %s",
                          type(g).__name__, str(g))
    return ad_groups



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from adles.utils import time_execution


# TODO: subclass this for platform-specific interfaces
class Interface:
    """ Generic interface used to uniformly interact with platform-specific interfaces. """

    __version__ = "1.0.1"  # This should match version of infrastructure-specification.yaml

    def __init__(self, spec, infra):
        """ 
        :param spec: Full exercise specification
        :param infra: Full infrastructure configuration
        """
        self._log = logging.getLogger('PlatformInterface')
        self.interfaces = []  # List of instantiated platform interfaces
        self.metadata = spec["metadata"]  # Save the exercise specification metadata
        self.infra = infra  # Save the infrastructure configuration

        # Select the Interface to use based on the specified infrastructure platform
        for platform, config in infra.items():
            if platform == "vmware-vsphere":
                from .vsphere_interface import VsphereInterface
                self.interfaces.append(VsphereInterface(config, spec))
            elif platform == "docker":
                from .docker_interface import DockerInterface
                self.interfaces.append(DockerInterface(config, spec))
            else:
                self._log.error("Invalid platform: %s", str(platform))
                raise ValueError

    @time_execution
    def create_masters(self):
        """ Master creation phase """
        self._log.info("Creating Master instances for %s", self.metadata["name"])
        # TODO: subprocess each interface call
        for i in self.interfaces:
            i.create_masters()

    @time_execution
    def deploy_environment(self):
        """ Environment deployment phase """
        self._log.info("Deploying environment for %s", self.metadata["name"])
        # TODO: subprocess each interface call
        for i in self.interfaces:
            i.deploy_environment()

    @time_execution
    def cleanup_masters(self, network_cleanup=False):
        """
        Cleans up master instances
        :param network_cleanup: If networks should be cleaned up [default: False]
        """
        self._log.info("Cleaning up Master instances for %s", self.metadata["name"])
        # TODO: subprocess each interface call
        for i in self.interfaces:
            i.cleanup_masters(network_cleanup=network_cleanup)

    @time_execution
    def cleanup_environment(self, network_cleanup=False):
        """
        Cleans up a deployed environment
        :param network_cleanup: If networks should be cleaned up [default: False]
        """
        self._log.info("Cleaning up environment for %s", self.metadata["name"])
        # TODO: subprocess each interface call
        for i in self.interfaces:
            i.cleanup_masters(network_cleanup=network_cleanup)

    def __repr__(self):
        return "Interface(%s,%s" % (str(self.interfaces), str(self.infra))

    def __str__(self):
        return str([x for x in self.infra.keys()])

    def __hash__(self):
        return hash(self.interfaces)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.interfaces == other.interfaces

    def __ne__(self, other):
        return not self.__eq__(other)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from os.path import exists

from netaddr import IPNetwork, AddrFormatError

import adles.utils as utils


# PyYAML Reference: http://pyyaml.org/wiki/PyYAMLDocumentation
def parse_file(filename):
    """
    Parses the YAML file and returns a nested dictionary containing it's contents.

    >>> parse_file("test.yaml")
    { "test": "value" }

    :param filename: Name of YAML file to parse
    :return: dictionary of parsed file contents
    """
    from yaml import load, YAMLError
    try:  # Attempt to use C-based YAML parser if it's available
        # noinspection PyUnresolvedReferences
        from yaml import CLoader as Loader
        logging.debug("Using C-based YAML parser")
    except ImportError:  # Fallback to using pure python YAML parser
        from yaml import Loader

    with open(filename) as f:
        try:
            doc = load(f, Loader=Loader)  # Parses the YAML file into a dict
        except YAMLError as exc:
            logging.error("Could not parse file %s", filename)
            if hasattr(exc, 'problem_mark'):
                mark = exc.problem_mark
                logging.error("Error position: (%s:%s)", mark.line + 1, mark.column + 1)
            else:
                logging.error("Error: %s", exc)
            return None
    return doc


def _checker(value_list, source, data, flag):
    """
    Checks if values in the list are in data (Syntax warnings or errors)
    :param value_list: List of values to check
    :param source: Name of source that's being checked
    :param data: Data being checked
    :param flag: What to do if value not found ("warnings" | "errors")
    :return: Number of hits (warnings/errors)
    """
    # TODO: add type checking, perhaps use a tuple of (value, type)
    num_hits = 0
    for value in value_list:
        if value not in data:
            if flag == "warnings":
                logging.warning("Missing %s in %s", value, source)
            elif flag == "errors":
                logging.error("Missing %s in %s", value, source)
            else:
                logging.error("Invalid flag for _checker: %s", flag)
            num_hits += 1
    if num_hits > 0:
        logging.info("Total number of %s in %s: %d", flag, source, num_hits)
    return num_hits


def _verify_exercise_metadata_syntax(metadata):
    """
    Verifies that the syntax for exercise metadata matches the specification
    :param metadata: Dict of metadata
    :return: (Number of errors, Number of warnings)
    """
    warnings = ["description", "version", "folder-name"]
    errors = ["name", "prefix", "infra-file"]

    num_warnings = _checker(warnings, "metadata", metadata, "warnings")
    num_errors = _checker(errors, "metadata", metadata, "errors")

    if "infra-file" in metadata:
        infra_file = metadata["infra-file"]
        if not exists(infra_file):
            logging.error("Could not open infra-file '%s'", infra_file)
            num_errors += 1
        else:
            e, w = verify_infra_syntax(parse_file(infra_file))
            num_errors += e
            num_warnings += w
    return num_errors, num_warnings


def _verify_groups_syntax(groups):
    """
    Verifies that the syntax for groups matches the specification
    :param groups: Dict of groups
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0
    
    for key, value in groups.items():
        if "instances" in value:  # Template groups
            if type(value["instances"]) != int:
                logging.error("Instances must be an Integer for group %s", key)
                num_errors += 1
            if "ad-group" in value:
                if type(value["ad-group"]) != str:
                    logging.error("AD group must be a string")
                    num_errors += 1
            elif "filename" in value:
                e, w = _check_group_file(value["filename"])
                num_errors += e
                num_warnings += w
            else:
                logging.error("Invalid user specification method for template group %s", key)
                num_errors += 1
        else:  # Regular groups (not templates)
            if "ad-group" in value:
                if type(value["ad-group"]) != str:
                    logging.error("AD group must be a string")
                    num_errors += 1
            elif "filename" in value:
                e, w = _check_group_file(value["filename"])
                num_errors += e
                num_warnings += w
            elif "user-list" in value:
                if type(value["user-list"]) is not list:
                    logging.error("Username specification must be a list for group %s", key)
                    num_errors += 1
            else:
                logging.error("Invalid user specification method for group %s", key)
                num_errors += 1
    return num_errors, num_warnings


def _check_group_file(filename):
    """
    Verifies user info file for a group
    :param filename: Name of user info JSON file
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0

    if utils.read_json(filename) is None:
        logging.error("Invalid user info file %s", filename)
        num_errors += 1
    return num_errors, num_warnings


def _verify_services_syntax(services):
    """
    Verifies that the syntax for services matches the specification
    :param services: Dict of services
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0
    
    for key, value in services.items():
        if "network-interfaces" in value and not isinstance(value["network-interfaces"], list):
            logging.error("Network interfaces must be a list for service %s", key)
            num_errors += 1
        if "provisioner" in value:
            if "name" not in value["provisioner"]:
                logging.error("Must specify name of provisioner for service %s", key)
                num_errors += 1
            if "file" not in value["provisioner"]:
                logging.error("Must specify provisioning file for service %s", key)
                num_errors += 1
        if "note" in value and not isinstance(value["note"], str):
            logging.error("Note must be a string for service %s", key)
            num_errors += 1
        if "template" in value:
            pass
        elif "image" in value or "dockerfile" in value:
            pass
        elif "compose-file" in value:
            pass
        else:
            logging.error("Invalid service definition: %s", key)
            num_errors += 1
    return num_errors, num_warnings


def _verify_resources_syntax(resources):
    """
    Verifies that the syntax for resources matches the specification
    :param resources: Dict of resources
    :return: (Number of errors, Number of warnings)
    """
    warnings = []
    errors = ["lab", "resource"]
    num_warnings = _checker(warnings, "resources", resources, "warnings")
    num_errors = _checker(errors, "resources", resources, "errors")
    return num_errors, num_warnings


def _verify_networks_syntax(networks):
    """
    Verifies that the syntax for networks matches the specification
    :param networks: Dict of networks
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0

    net_types = ["unique-networks", "generic-networks"]
    if not any(net in networks for net in net_types):
        logging.error("Network specification exists but is empty!")
        num_errors += 1
    else:
        for name, network in networks.items():
            e, w = _verify_network(name, network)
            num_errors += e
            num_warnings += w
    return num_errors, num_warnings


def _verify_network(name, network):
    """
    Verifies syntax of a specific network
    :param name: Name of network
    :param network: Dict of the network
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0
    
    for key, value in network.items():
        # Subnet verification
        if "subnet" not in value:
            logging.warning("No subnet specified for %s %s", name, key)
            num_warnings += 1
        else:
            try:
                subnet = IPNetwork(value["subnet"])
            except AddrFormatError:
                logging.error("Invalid format for subnet '%s'", str(value["subnet"]))
                num_errors += 1
            else:
                if subnet.is_reserved() or subnet.is_multicast() or subnet.is_loopback():
                    logging.error("%s %s is in a invalid IP address space", name, key)
                    num_errors += 1
                elif not subnet.is_private():
                    logging.warning("Non-private subnet used for %s %s", name, key)
                    num_warnings += 1

        # VLAN verification
        if "vlan" in value:
            if name == "unique-networks" and int(value["vlan"]) >= 2000:
                logging.error("VLAN must be less than 2000 for network %s", key)
                num_errors += 1
            elif name == "generic-networks":
                logging.error("VLAN specification is not allowed for network %s", key)
                num_errors += 1

        # Increment verification
        if "increment" in value:
            if name == "unique-networks":
                logging.error("Increment cannot be used for network %s", key)
                num_errors += 1
            elif type(value["increment"]) != bool:
                logging.error("Increment must be a boolean for network %s", key)
                num_errors += 1
    return num_errors, num_warnings


def _verify_folders_syntax(folders):
    """
    Verifies that the syntax for folders matches the specification
    :param folders: Dict of folders
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0
    keywords = ["group", "master-group", "instances", "description", "enabled"]

    for key, value in folders.items():
        if key in keywords:
            continue
        if type(value) is not dict:
            logging.error("Invalid configuration %s", str(key))
            num_errors += 1
            continue
        if "instances" in value:  # Check instances syntax, regardless of parent or base
            if type(value["instances"]) is int:
                pass
            elif "number" in value["instances"]:
                if type(value["instances"]["number"]) != int:
                    logging.error("Number of instances for folder '%s' must be an Integer", key)
                    num_errors += 1
            elif "size-of" in value["instances"]:
                pass  # TODO: verify group existence
            else:
                logging.error("Must specify number of instances for folder '%s'", key)
                num_errors += 1

        # Check if parent or base
        if "services" in value:  # It's a base folder
            if "group" not in value:
                logging.error("No group specified for folder '%s'", key)
                num_errors += 1
            for skey, svalue in value["services"].items():
                if "service" not in svalue:
                    logging.error("Service %s is unnamed in folder '%s'", skey, key)
                    num_errors += 1
                if "networks" in svalue and type(svalue["networks"]) is not list:
                    logging.error("Network specifications must be a list for service '%s' "
                                  "in folder '%s'", skey, key)
                    num_errors += 1
                if "scoring" in svalue:
                    e, w = _verify_scoring_syntax(skey, svalue["scoring"])
                    num_errors += e
                    num_warnings += w
        else:  # It's a parent folder
            if not isinstance(value, dict):
                logging.error("Could not verify syntax of folder '%s': '%s' is not a folder!",
                              str(key), str(value))
                num_errors += 1
            else:
                e, w = _verify_folders_syntax(value)
                num_errors += e
                num_warnings += w
    return num_errors, num_warnings


def _verify_scoring_syntax(service_name, scoring):
    """
    Verifies syntax for the scoring definition of a service.
    :param service_name: Name of the service for which the scoring specification applies
    :param scoring: Dict of scoring parameters
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0

    if "ports" not in scoring:
        logging.warning("No scoring ports specified for service %s", service_name)
        num_warnings += 1
    elif not isinstance(scoring["ports"], list):
        logging.error("Scoring ports must be a list for service %s", service_name)
        num_errors += 1
    if "protocols" not in scoring:
        logging.warning("No protocols specified for service %s", service_name)
        num_warnings += 1
    elif not isinstance(scoring["protocols"], list):
        logging.error("Protocols must be a list for service %s", service_name)
        num_errors += 1
    if "criteria" not in scoring:
        logging.error("No criteria file specified for service %s", service_name)
        num_errors += 1
    return num_errors, num_warnings


def verify_infra_syntax(infra):
    """
    Verifies the syntax of an infrastructure specification.
    :param infra: Dict of infrastructure
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0
    warnings = []
    errors = []

    for platform, config in infra.items():
        if platform == "vmware-vsphere":  # VMware vSphere configurations
            warnings = ["port", "login-file", "datacenter", "datastore", "server-root", "vswitch"]
            errors = ["hostname", "template-folder"]
            if "login-file" in config and utils.read_json(config["login-file"]) is None:
                logging.error("Invalid vSphere infrastructure login-file: %s", config["login-file"])
                num_errors += 1
            if "host-list" in config and type(config["host-list"]) is not list:
                logging.error("Invalid type for vSphere host-list: %s", type(config["host-list"]))
                num_errors += 1
        elif platform == "docker":  # Docker configurations
            warnings = ["url"]
            errors = []
            if "registry" in config:
                num_errors += _checker(["url", "login-file"], "infrastructure",
                                       config["registry"], "errors")
        elif platform == "amazon-aws":
            pass
        elif platform == "digital-ocean":
            pass
        elif platform == "hyper-v":
            pass
        else:
            logging.error("Unknown infrastructure platform: %s", str(platform))
            continue  # Skip the syntax verification of unknown platforms
        num_warnings += _checker(warnings, "infrastructure", config, "warnings")
        num_errors += _checker(errors, "infrastructure", config, "errors")
    return num_errors, num_warnings


def verify_exercise_syntax(spec):
    """
    Verifies the syntax of an environment specification.
    :param spec: Dictionary of environment specification
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0
    funcs = {"metadata": _verify_exercise_metadata_syntax,
             "groups": _verify_groups_syntax,
             "services": _verify_services_syntax,
             "resources": _verify_resources_syntax,
             "networks": _verify_networks_syntax,
             "folders": _verify_folders_syntax}

    required = ["metadata", "groups", "services", "networks", "folders"]
    optional = ["resources"]

    for key, func in funcs.items():
        if key not in spec:
            if key in required:
                logging.error("Required definition %s was not found", key)
                num_errors += 1
            elif key in optional:
                logging.debug('Optional definition "%s" was not found', key)
            else:
                logging.warning("Unknown definition found: %s", key)
                num_warnings += 1
        else:
            e, w = func(spec[key])
            num_errors += e
            num_warnings += w
    return num_errors, num_warnings


def verify_package_syntax(package):
    """
    Verifies the syntax of an package specification.
    :param package: Dictionary representation of the package specification
    :return: (Number of errors, Number of warnings)
    """
    num_warnings = 0
    num_errors = 0

    # Check syntax of metadata section
    if "metadata" not in package:
        logging.error("Metadata section not specified for package!")
        num_errors += 1
    else:
        metadata_warnings = ["name", "description", "version"]
        metadata_errors = ["timestamp", "tag"]
        num_warnings += _checker(metadata_warnings, "metadata", package["metadata"], "warnings")
        num_errors += _checker(metadata_errors, "metadata", package["metadata"], "errors")

    # Check syntax of contents section
    if "contents" not in package:
        logging.error("Contents section not specified for package!")
        num_errors += 1
    else:
        content_warnings = ["infrastructure", "scoring", "results", "templates", "materials"]
        content_errors = ["environment"]
        num_warnings += _checker(content_warnings, "contents", package["contents"], "warnings")
        num_errors += _checker(content_errors, "contents", package["contents"], "errors")

    return num_errors, num_warnings


def check_syntax(specfile_path, spec_type="exercise"):
    """
    Checks the syntax of a specification file
    :param specfile_path: Path to the specification file
    :param spec_type: Type of specification file (exercise | package | infra)
    :return: The specification
    """
    from os.path import exists, basename

    if not exists(specfile_path):
        logging.error("Could not find specification file in path %s", str(specfile_path))
        return None
    spec = parse_file(specfile_path)
    if spec is None:
        logging.error("Failed to ingest specification file %s", basename(specfile_path))
        return None
    logging.info("Successfully ingested specification file %s", basename(specfile_path))
    logging.info("Checking syntax...")
    if spec_type == "exercise":
        errors, warnings = verify_exercise_syntax(spec)
    elif spec_type == "package":
        errors, warnings = verify_package_syntax(spec)
    elif spec_type == "infra":
        errors, warnings = verify_infra_syntax(spec)
    else:
        logging.error("Unknown specification type in for check_syntax: %s", str(spec_type))
        return None

    if errors == 0 and warnings == 0:
        logging.info("Syntax check successful!")
        return spec
    elif errors == 0:
        logging.info("Syntax check successful, but there were %d warnings", warnings)
        return spec
    else:
        logging.error("Syntax check failed! Errors: %d\tWarnings: %d", errors, warnings)
        return None



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import logging.handlers

from sys import stdout, exit
import os


def check(arg_type, kwarg_name):
    """
    Function decorator that validates types of parameters
    :param arg_type: Type to validate
    :param kwarg_name: Name of keyword argument to validate
    :return: The decorated function
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            if args and isinstance(args[0], arg_type):  # We assume it is the first argument
                return func(*args, **kwargs)
            elif kwargs and isinstance(kwargs.get(kwarg_name, None), arg_type):
                return func(*args, **kwargs)
            else:
                logging.error("Function '%s' failed check for type '%s'\nArgs: %s\nkwargs: %s",
                              func.__name__, arg_type.__name__, str(args), str(kwargs))
        return wrapper
    return decorator


# Credit to: http://stackoverflow.com/a/15707426/2214380
def time_execution(func):
    """
    Function decorator to time the execution of a function and log to debug
    :param func: The function to time execution of
    :return: The decorated function
    """
    from timeit import default_timer

    def wrapper(*args, **kwargs):
        start_time = default_timer()
        ret = func(*args, **kwargs)
        end_time = default_timer()
        logging.debug("Elapsed time for %s: %f seconds", func.__name__,
                      float(end_time - start_time))
        return ret
    return wrapper


# From: list_dc_datastore_info in pyvmomi-community-samples
# http://stackoverflow.com/questions/1094841/
def sizeof_fmt(num):
    """
    Generates the human-readable version of a file size

    >>> sizeof_fmt(512)
    512bytes
    >>> sizeof_fmt(2048)
    2KB

    :param num: Robot-readable file size in bytes
    :return: Human-readable file size
    """
    for item in ['bytes', 'KB', 'MB', 'GB']:
        if num < 1024.0:
            return "%3.1f%s" % (num, item)
        num /= 1024.0
    return "%3.1f%s" % (num, 'TB')


# From: virtual_machine_power_cycle_and_question in pyvmomi-community-samples
def _create_char_spinner():
    """ Creates a generator yielding a char based spinner """
    while True:
        for c in '|/-\\':
            yield c

_spinner = _create_char_spinner()


def spinner(label=""):
    """
    When called repeatedly from inside a loop this prints a CLI spinner
    :param label: The message to display while spinning [default: ""]
    """
    stdout.write("\r\t%s %s" % (label, next(_spinner)))
    stdout.flush()


def pad(value, length=2):
    """
    Adds leading and trailing zeros to value ("pads" the value)

    >>> pad(5)
    05
    >>> pad(9, 3)
    009

    :param value: integer value to pad
    :param length: Length to pad to [default: 2]
    :return: string of padded value
    """
    return "{0:0>{width}}".format(value, width=length)


def read_json(filename):
    """
    Reads input from a JSON file and returns the contents
    :param filename: Path to JSON file to read
    :return: Contents of the JSON file
    """
    from json import load
    try:
        with open(filename) as json_file:
            return load(fp=json_file)
    except ValueError as e:
        logging.error("Syntax Error in JSON file '%s': %s", filename, str(e))
        return None
    except Exception as e:
        logging.error("Could not open JSON file '%s': %s", filename, str(e))
        return None


def split_path(path):
    """
    
    :param path: 
    :return: (Path, basename)
    """
    folder_path, name = os.path.split(path.lower())  # Separate basename and convert to lowercase
    folder_path = folder_path.split('/')  # Transform path into list
    if folder_path[0] == '':
        del folder_path[0]
    return folder_path, name


def make_vsphere(filename=None):
    """
    Creates a vSphere object using either a JSON file or by prompting the user
    :param filename: Name of JSON file with information needed [default: None]
    :return: vSphere object
    """
    from adles.vsphere.vsphere_class import Vsphere

    if filename is not None:
        info = read_json(filename)
        return Vsphere(username=info.get("user"),
                       password=info.get("pass"),
                       hostname=info.get("host"),
                       port=info.get("port", 443),
                       datacenter=info.get("datacenter"),
                       datastore=info.get("datastore"))
    else:
        logging.info("Enter information to connect to vSphere environment")
        datacenter = input("Datacenter  : ")
        datastore = input("Datastore   : ")
        return Vsphere(datacenter=datacenter, datastore=datastore)


def user_input(prompt, obj_name, func):
    """
    Continually prompts a user for input until the specified object is found
    :param prompt: Prompt to bother user with
    :param obj_name: Name of the type of the object that we seek
    :param func: The function that shalt be called to discover the object
    :return: The discovered object and it's human name
    """
    while True:
        try:
            item_name = str(input(prompt))
        except KeyboardInterrupt:
            print()
            logging.info("Exiting...")
            exit(0)
        item = func(item_name)
        if item:
            logging.info("Found %s: %s", obj_name, item.name)
            return item, item_name
        else:
            print("Couldn't find a %s with name %s. Perhaps try another? "
                  % (obj_name, item_name))


# Based on: http://code.activestate.com/recipes/577058/
def prompt_y_n_question(question, default="no"):
    """
    Prompts user to answer a question
    :param question: Question to ask
    :param default: No
    :return: True/False
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    choice = ''
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("Invalid default answer: '%s'", default)

    while True:
        try:
            choice = str(input(question + prompt)).lower()
        except KeyboardInterrupt:
            print()  # Output a blank line for readability
            logging.info("Exiting...")
            exit(0)

        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' or 'y' or 'n'")


def default_prompt(prompt, default=None):
    """
    Prompt the user for input. If they press enter, return the default.
    :param prompt: Prompt to display to user (do not include default value)
    :param default: Default return value
    :return: Value returned
    """
    try:
        value = str(input(prompt + " [default: %s]: " % str(default)))
    except KeyboardInterrupt:
        print()  # Output a blank line for readability
        logging.info("Exiting...")
        exit(0)

    # noinspection PyUnboundLocalVariable
    if value == '':
        return default
    else:
        return value


def script_warning_prompt():
    """
    Generates a warning prompt
    :return: The warning prompt
    """
    from adles import __url__, __email__
    return str(
        '***** YOU RUN THIS SCRIPT AT YOUR OWN RISK *****\n'
        '\nGetting help:'
        '\n\t* "<script>.py --help": flags, arguments, and usage'
        '\n\t* "cat <script>.py": read the source code and see how it works'
        '\n\t* "cd ./documentation && ls -la": show available documentation'
        '\n\t+ Open an issue on the project GitHub: %s'
        '\n\t+ Email the script author: %s'
        '\n\n' % (__url__, __email__))


def script_setup(logging_filename, args, script=None):
    """
    Does setup tasks that are common to all automation scripts
    :param logging_filename: Name of file to save logs to
    :param args: docopt arguments dict
    :param script: Tuple with name and version of the script [default: None]
    :return: vSphere object
    """

    # Setup logging
    colors = (False if args["--no-color"] else True)
    setup_logging(filename=logging_filename, colors=colors,
                  console_verbose=args["--verbose"])

    # Print information about script itself
    if script:
        logging.debug("Script name      %s", os.path.basename(script[0]))
        logging.debug("Script version   %s", script[1])
        print(script_warning_prompt())  # Print warning for script users

    # Create the vsphere object and return it
    try:
        return make_vsphere(args["--file"])
    except KeyboardInterrupt:
        print()  # Output a blank line for readability
        logging.info("Exiting...")
        exit(0)


def resolve_path(server, thing, prompt=""):
    """
    This is a hacked together script util to get folders or VMs
    :param server: Vsphere instance
    :param thing: String name of thing to get (folder | vm)
    :param prompt: Message to display [default: ""]
    :return: (thing, thing name)
    """
    from adles.vsphere.folder_utils import traverse_path
    if thing.lower() == "vm":
        get = server.get_vm
    elif thing.lower() == "folder":
        get = server.get_folder
    else:
        logging.error("Invalid thing passed to resolve_path: %s", thing)
        raise ValueError

    return user_input("Name of or path to %s %s: " % (thing, prompt), thing,
                      lambda x: traverse_path(server.get_folder(), x, lookup_root=server)
                      if '/' in x else get(x))


def setup_logging(filename, colors=True, console_verbose=False,
                  server=('localhost', 514)):
    """
    Configures the logging interface used by everything for output
    :param filename: Name of file that logs should be saved to
    :param colors: Color the terminal output [default: True]
    :param console_verbose: Print DEBUG logs to terminal [default: False]
    :param server: SysLog server to forward logs to [default: (localhost, 514)]
    """

    # Prepend spaces to separate logs from previous runs
    with open(filename, 'a') as logfile:
        logfile.write(2 * '\n')

    # Format log output so it's human readable yet verbose
    base_format = "[%(asctime)s] %(name)-8s %(levelname)-8s %(message)s"
    time_format = "%Y-%m-%d %H:%M:%S"
    formatter = logging.Formatter(fmt=base_format, datefmt=time_format)

    # Configures the base logger to append to a file
    logging.basicConfig(level=logging.DEBUG, format=base_format, datefmt=time_format,
                        filename=filename, filemode='a')

    # Get the global root logger
    logger = logging.root

    # Configure logging to a SysLog server
    # This prevents students from simply deleting the log files
    syslog = logging.handlers.SysLogHandler(address=server)
    syslog.setLevel(logging.DEBUG)
    syslog.setFormatter(formatter)
    logger.addHandler(syslog)
    logging.debug("Configured logging to SysLog server %s:%s", server[0], str(server[1]))

    # TODO: configure logging to Splunk (https://github.com/zach-taylor/splunk_handler)

    # Configure console output
    console = logging.StreamHandler(stream=stdout)
    if colors:  # Colored console output
        try:
            # noinspection PyUnresolvedReferences
            from colorlog import ColoredFormatter
            formatter = ColoredFormatter(fmt="%(log_color)s" + base_format,
                                         datefmt=time_format, reset=True)
            logging.debug("Configured COLORED console logging output")
        except ImportError:
            logging.error("Colorlog is not installed. Using STANDARD console output...")
    else:  # Bland console output
        logging.debug("Configured STANDARD console logging output")
    console.setFormatter(formatter)
    console.setLevel((logging.DEBUG if console_verbose else logging.INFO))
    logger.addHandler(console)

    # Record system information to aid in auditing and debugging
    from getpass import getuser
    from os import getcwd
    from platform import python_version, platform
    from adles import __version__ as adles_version
    from adles.vsphere.vsphere_class import Vsphere
    logging.debug("Initialized logging, saving logs to %s", filename)
    logging.debug("Python           %s", str(python_version()))
    logging.debug("Platform         %s", str(platform()))
    logging.debug("Username         %s", str(getuser()))
    logging.debug("Directory        %s", str(getcwd()))
    logging.debug("Adles version    %s", str(adles_version))
    logging.debug("Vsphere version  %s", str(Vsphere.__version__))



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from pyVmomi import vim

import adles.utils as utils
from adles.vsphere.vsphere_utils import wait_for_task


class Folder:
    """ Represents a VMware vSphere Folder instance. """
    __version__ = "0.1.0"

    def __init__(self, folder, host):
        self._log = logging.getLogger('Folder')
        self.folder = folder
        self.name = folder.name
        self.host = host

    def move_into(self, entity_list):
        """
        Moves a list of managed entities into the named folder.
        :param folder: vim.Folder object with type matching the entity list
        :param entity_list: List of vim.ManagedEntity
        """
        wait_for_task(self.folder.MoveIntoFolder_Task(entity_list))

    def __str__(self):
        return str(self.name)

    def __hash__(self):
        return hash(self.folder)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.name == other.name

    def __ne__(self, other):
        return not self.__eq__(other)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from pyVmomi import vim

import adles.utils as utils
from adles.vsphere.vsphere_utils import wait_for_task, is_vnic
from adles.vsphere.folder_utils import find_in_folder


class VM:
    """ Represents a VMware vSphere Virtual Machine instance. """
    __version__ = "0.1.0"

    def __init__(self, name, folder, resource_pool, datastore, host):
        """
        NOTE: VM.create() must be called post-init
        :param name: Name of the VM
        :param folder: Name of the folder VM is located
        :param resource_pool: Resource pool to use for the VM
        :param datastore: Datastore the VM is stored on
        :param host: Host the VM runs on
        """
        self._log = logging.getLogger('VM')
        self.name = str(name)
        self.folder = folder    # vim.Folder, but may be Folder class eventually
        self.resource_pool = resource_pool  # vim.Pool or something object
        self.datastore = datastore      # vim.Datastore object
        self.host = host        # vim.HostSystem, but may be Host class soon
        self._vm = None

    def create(self, template=None, cpus=1, cores=1, memory=512, max_consoles=None,
               version=None, firmware='efi', datastore_path=None):
        """
        Creates a Virtual Machine
        :param template: Template VM to clone
        :param cpus: Number of processors [default: 1]
        :param cores: Number of processor cores [default: 1]
        :param memory: Amount of RAM in MB [default: 512]
        :param max_consoles: Maximum number of active console connections [default: default]
        :param version: Hardware version of the VM [default: highest host supports]
        :param firmware: Firmware to emulate for the VM (efi | bios) [default: efi]
        :param datastore_path: Path to existing VM files on datastore [default: None]
        """
        # Use a template to create the VM
        if template is not None:
            clonespec = vim.vm.CloneSpec()
            clonespec.location = vim.vm.RelocateSpec(pool=self.resource_pool,
                                                     datastore=self.datastore)
            try:
                wait_for_task(template.CloneVM_Task(folder=self.folder, name=self.name,
                                                    spec=clonespec))
            except vim.fault.InvalidState:
                logging.error("Could not make clone '%s': invalid state for VM '%s'", self.name,
                              template.name)
            except vim.fault.CustomizationFault:
                logging.error("Could not make clone '%s': invalid customization", self.name)
            except vim.fault.VmConfigFault:
                logging.error("Could not make clone '%s': invalid configuration", self.name)
        else:
            # Generate the specification for the new VM
            spec = vim.vm.ConfigSpec()
            spec.name = self.name
            spec.numCPUs = int(cpus)
            spec.numCoresPerSocket = int(cores)
            spec.memoryMB = int(memory)
            spec.memoryHotAddEnabled = True
            spec.firmware = str(firmware).lower()
            if version is not None:
                spec.version = str(version)
            if max_consoles is not None:
                spec.maxMksConnections = int(max_consoles)
            vm_path = '[' + self.datastore.name + '] '
            if datastore_path:
                vm_path += str(datastore_path)
            vm_path += self.name + '/' + self.name + '.vmx'
            spec.files = vim.vm.FileInfo(vmPathName=vm_path)
            self._log.debug("Creating VM '%s' in folder '%s'", self.name, self.folder.name)
            wait_for_task(self.folder.CreateVM_Task(spec, self.resource_pool, self.host))
        self._vm = find_in_folder(self.folder, self.name, vimtype=vim.VirtualMachine)

    def destroy(self):
        """ Destroy the VM """
        self._log.debug("Destroying VM %s", self.name)
        if self.powered_on():
            self.change_state("off")
        wait_for_task(self._vm.Destroy_Task())

    def change_state(self, state, attempt_guest=True):
        """
        Generic power state change function that uses guest operations if available
        :param state: State to change to (on | off | reset | suspend)
        :param attempt_guest: Whether to attempt to use guest operations to change power state
        :return: 
        """
        state = state.lower()
        if self.is_template():
            logging.error("VM '%s' is a Template, so state cannot be changed to '%s'", self.name,
                          state)
        elif attempt_guest and self.has_tools() and state != "on":  # Can't power on using guest ops
            if self._vm.summary.guest.toolsStatus == "toolsNotInstalled":
                logging.error("Cannot change a VM's guest power state without VMware Tools!")
                return
            elif state == "shutdown" or state == "off":
                task = self._vm.ShutdownGuest()
            elif state == "reboot" or state == "reset":
                task = self._vm.RebootGuest()
            elif state == "standby" or state == "suspend":
                task = self._vm.StandbyGuest()
            else:
                logging.error("Invalid guest_state argument: %s", state)
                return
            logging.debug("Changing guest power state of VM %s to: '%s'", self.name, state)
            try:
                wait_for_task(task)
            except vim.fault.ToolsUnavailable:
                logging.error("Cannot change guest state of '%s': Tools are not running", self.name)
        else:
            if state == "on":
                task = self._vm.PowerOnVM_Task()
            elif state == "off":
                task = self._vm.PowerOffVM_Task()
            elif state == "reset":
                task = self._vm.ResetVM_Task()
            elif state == "suspend":
                task = self._vm.SuspendVM_Task()
            else:
                logging.error("Invalid state arg %s for VM %s", state, self.name)
                return
            logging.debug("Changing power state of VM %s to: '%s'", self.name, state)
            wait_for_task(task)

    def upgrade_vm(self, version):
        """
        Upgrades the hardware version of the VM
        :param version: Version of hardware to upgrade VM to [default: latest VM's host supports]
        """
        try:
            wait_for_task(self._vm.UpgradeVM(str(version)))
        except vim.fault.AlreadyUpgraded:
            logging.error("Hardware version is already up-to-date for %s", self.name)
        except vim.fault.InvalidPowerState as e:
            logging.error("Cannot upgrade '%s': invalid power state '%s'", self.name,
                          e.existingState)

    def convert_template(self):
        """
        Converts a Virtual Machine to a template
        :return: 
        """
        if self.is_template():
            self._log.debug("%s is already a template", self.name)
        else:
            self._log.debug("Converting %s to template", self.name)
            self._vm.MarkAsTemplate()

    def convert_vm(self, resource_pool, host=None):
        """
        Converts a Template to a Virtual Machine
        :param resource_pool: vim.ResourcePool to associate with the VM
        :param host: vim.HostSystem on which the VM should run [default: None]
        """
        logging.debug("Converting '%s' to VM and assigning to resource pool '%s'",
                      self.name, resource_pool.name)
        self._vm.MarkAsVirtualMachine(resource_pool, host)

    def set_note(self, note):
        """
        Sets the note on the VM to note
        :param note: String to set the note to
        """
        spec = vim.vm.ConfigSpec()
        spec.annotation = note
        self._edit(spec)

    def get_vm_info(self, detailed=False, uuids=False, snapshot=False, vnics=False):
        """
        Get human-readable information for a VM
        :param detailed: Add more detailed information, such as maximum memory used [default: False]
        :param uuids: Whether to get UUID information [default: False]
        :param snapshot: Shows the current snapshot, if any [default: False]
        :param vnics: Add information about the virtual network interfaces on the VM
        :return: String with the VM information
        """
        info_string = "\n"
        summary = self._vm.summary
        info_string += "Name          : %s\n" % summary.config.name
        info_string += "Status        : %s\n" % str(summary.overallStatus)
        info_string += "Power State   : %s\n" % summary.runtime.powerState
        if self._vm.guest:
            info_string += "Guest State   : %s\n" % self._vm.guest.guestState
        info_string += "Last modified : %s\n" % str(self._vm.config.modified)  # datetime object
        if detailed:
            info_string += "Num consoles  : %d\n" % summary.runtime.numMksConnections
        info_string += "Host          : %s\n" % summary.runtime.host.name
        info_string += "Guest OS      : %s\n" % summary.config.guestFullName
        info_string += "Num CPUs      : %s\n" % summary.config.numCpu
        info_string += "Memory (MB)   : %s\n" % summary.config.memorySizeMB
        if detailed:
            info_string += "Num vNICs     : %s\n" % summary.config.numEthernetCards
            info_string += "Num Disks     : %s\n" % summary.config.numVirtualDisks
        info_string += "IsTemplate    : %s\n" % summary.config.template  # bool
        if detailed:
            info_string += "Path          : %s\n" % summary.config.vmPathName
        info_string += "Folder:       : %s\n" % self._vm.parent.name
        if self._vm.guest:
            info_string += "IP            : %s\n" % self._vm.guest.ipAddress
            info_string += "Hostname:     : %s\n" % self._vm.guest.hostName
            info_string += "Tools status  : %s\n" % self._vm.guest.toolsRunningStatus
            info_string += "Tools version : %s\n" % self._vm.guest.toolsVersionStatus2
        if vnics:
            vm_nics = self.get_nics()
            for num, vnic in zip(range(1, len(vm_nics) + 1), vm_nics):
                info_string += "vNIC %d label   : %s\n" % (num, vnic.deviceInfo.label)
                info_string += "vNIC %d summary : %s\n" % (num, vnic.deviceInfo.summary)
                info_string += "vNIC %d network : %s\n" % (num, vnic.backing.network.name)
        if uuids:
            info_string += "Instance UUID : %s\n" % summary.config.instanceUuid
            info_string += "Bios UUID     : %s\n" % summary.config.uuid
        if summary.runtime.question:
            info_string += "Question      : %s\n" % summary.runtime.question.text
        if summary.config.annotation:
            info_string += "Annotation    : %s\n" % summary.config.annotation
        if snapshot and self._vm.snapshot and hasattr(self._vm.snapshot, 'currentSnapshot'):
            info_string += "Current Snapshot: %s\n" % self._vm.snapshot.currentSnapshot.config.name
            info_string += "Disk usage of all snapshots: %s\n" % self.snapshot_disk_usage()
        if detailed and summary.runtime:
            info_string += "Last Poweron  : %s\n" % str(summary.runtime.bootTime)  # datetime object
            info_string += "Max CPU usage : %s\n" % summary.runtime.maxCpuUsage
            info_string += "Max Mem usage : %s\n" % summary.runtime.maxMemoryUsage
            info_string += "Last suspended: %s\n" % summary.runtime.suspendTime
        return info_string

    def snapshot_disk_usage(self):
        """
        Determines the total disk usage of a VM's snapshots
        :return: Human-readable disk usage of the snapshots
        """
        from re import search
        disk_list = self._vm.layoutEx.file
        size = 0
        for disk in disk_list:
            if disk.type == 'snapshotData':
                size += disk.size
            ss_disk = search('0000\d\d', disk.name)
            if ss_disk:
                size += disk.size
        return utils.sizeof_fmt(size)

    def get_snapshot(self, snapshot=None):
        """
        Retrieves the named snapshot from the VM
        :param snapshot: Name of the snapshot to get [default: current snapshot]
        :return: vim.Snapshot object
        """
        if snapshot is None:
            return self._vm.snapshot.currentSnapshot
        else:
            for snap in self.get_all_snapshots():
                if snap.name == snapshot:
                    return snap.snapshot
            return None

    def get_all_snapshots(self):
        """
        Retrieves a list of all snapshots of the VM
        :return: Nested List of vim.Snapshot objects
        """
        return self._get_snapshots_recursive(self._vm.snapshot.rootSnapshotList)

    # From: https://github.com/imsweb/ezmomi
    def _get_snapshots_recursive(self, snap_tree):
        """
        Recursively finds all snapshots in the tree
        :param snap_tree: Tree of snapshots
        :return: Nested List of vim.Snapshot objects
        """
        local_snap = []
        for snap in snap_tree:
            local_snap.append(snap)
        for snap in snap_tree:
            recurse_snap = self._get_snapshots_recursive(snap.childSnapshotList)
            if recurse_snap:
                local_snap.extend(recurse_snap)
        return local_snap

    def create_snapshot(self, name, description='default', memory=False):
        """
        Create a snapshot of the VM
        :param name: Title of the snapshot
        :param memory: Memory dump of the VM is included in the snapshot
        :param description: Text description of the snapshot
        """
        logging.info("Creating snapshot '%s' of '%s'", name, self.name)
        wait_for_task(self._vm.CreateSnapshot_Task(name=name, description=description,
                                                   memory=bool(memory), quiesce=True))

    def revert_to_snapshot(self, snapshot):
        """
        Reverts VM to the named snapshot
        :param snapshot: Name of the snapshot to revert to
        """
        logging.info("Reverting '%s' to the snapshot '%s'", self.name, snapshot)
        snap = self.get_snapshot(snapshot)
        wait_for_task(snap.RevertToSnapshot_Task())

    def revert_to_current_snapshot(self):
        """ Reverts the VM to the most recent snapshot. """
        logging.info("Reverting '%s' to the current snapshot", self.name)
        wait_for_task(self._vm.RevertToCurrentSnapshot_Task())

    def remove_snapshot(self, snapshot_name, remove_children=True, consolidate_disks=True):
        """
        Removes the named snapshot from the VM
        :param snapshot_name: Name of the snapshot to remove
        :param remove_children: Removal of the entire snapshot subtree [default: True]
        :param consolidate_disks: Virtual disks of deleted snapshot will be merged with
        other disks if possible [default: True]
        """
        logging.info("Removing snapshot '%s' from '%s'", snapshot_name, self.name)
        snapshot = self.get_snapshot(snapshot_name)
        wait_for_task(snapshot.RemoveSnapshot_Task(remove_children, consolidate_disks))

    def remove_all_snapshots(self, consolidate_disks=True):
        """
        Removes all snapshots associated with the VM
        :param consolidate_disks: Virtual disks of the deleted snapshot will be merged with
        other disks if possible [default: True]
        """
        logging.info("Removing ALL snapshots for the '%s'", self.name)
        wait_for_task(self._vm.RemoveAllSnapshots_Task(consolidate_disks))

    def remove_device(self, device):
        """
        Removes the device from the VM
        :param device: vim.vm.device.VirtualDeviceSpec
        """
        logging.debug("Removing device '%s' from '%s'", device.name, self.name)
        device.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
        self._edit(vim.vm.ConfigSpec(deviceChange=[device]))

    def get_nics(self):
        """
        Returns a list of all Virtual Network Interface Cards (vNICs) on a VM
        :return: list of vim.vm.device.VirtualEthernetCard
        """
        return [dev for dev in self._vm.config.hardware.device if is_vnic(dev)]

    def get_nic_by_name(self, name):
        """
        Gets a Virtual Network Interface Card from a VM
        :param name: Name of the vNIC
        :return: vim.vm.device.VirtualEthernetCard
        """
        for dev in self._vm.config.hardware.device:
            if is_vnic(dev) and dev.deviceInfo.label.lower() == name.lower():
                return dev
        logging.debug("Could not find vNIC '%s' on '%s'", name, self.name)
        return None

    def get_nic_by_id(self, nic_id):
        """
        Get a vNIC by integer ID
        :param nic_id: ID of the vNIC
        :return: vim.vm.device.VirtualEthernetCard
        """
        return self.get_nic_by_name("Network Adapter " + str(nic_id))

    def get_nic_by_network(self, network):
        """
        Finds a vNIC by it's network backing
        :param network: vim.Network
        :return: Name of the vNIC
        """
        for dev in self._vm.config.hardware.device:
            if is_vnic(dev) and dev.backing.network == network:
                return dev
        logging.debug("Could not find vNIC with network '%s' on '%s'", network.name, self.name)
        return None

    def add_nic(self, network, summary="default-summary", model="e1000"):
        """
        Add a NIC in the portgroup to the VM
        :param network: vim.Network to attach NIC to
        :param summary: Human-readable device info [default: default-summary]
        :param model: Model of virtual network adapter. [default: e1000]
        Options: (e1000 | e1000e | vmxnet | vmxnet2 | vmxnet3)
        e1000 will work on Windows Server 2003+, and e1000e is supported on Windows Server 2012+.
        VMXNET adapters require VMware Tools to be installed, and provide enhanced performance.
        Read this for more details: http://rickardnobel.se/vmxnet3-vs-e1000e-and-e1000-part-1/
        """
        if not isinstance(network, vim.Network):
            logging.error("Invalid network type when adding vNIC to VM '%s': %s",
                          self.name, type(network).__name__)
        logging.debug("Adding NIC to VM '%s'\nNetwork: '%s'\tSummary: '%s'\tNIC Model: '%s'",
                      self.name, network.name, summary, model)
        nic_spec = vim.vm.device.VirtualDeviceSpec()  # Create base object to add configurations to
        nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add

        # Set the type of network adapter
        if model == "e1000":
            nic_spec.device = vim.vm.device.VirtualE1000()
        elif model == "e1000e":
            nic_spec.device = vim.vm.device.VirtualE1000e()
        elif model == "vmxnet":
            nic_spec.device = vim.vm.device.VirtualVmxnet()
        elif model == "vmxnet2":
            nic_spec.device = vim.vm.device.VirtualVmxnet2()
        elif model == "vmxnet3":
            nic_spec.device = vim.vm.device.VirtualVmxnet3()
        else:
            logging.error("Invalid NIC model: '%s'\nDefaulting to e1000...", model)
            nic_spec.device = vim.vm.device.VirtualE1000()
        nic_spec.device.addressType = 'generated'  # Sets how MAC address is assigned
        nic_spec.device.wakeOnLanEnabled = False  # Disables Wake-on-lan capabilities

        nic_spec.device.deviceInfo = vim.Description()
        nic_spec.device.deviceInfo.summary = summary

        nic_spec.device.backing = vim.vm.device.VirtualEthernetCard.NetworkBackingInfo()
        nic_spec.device.backing.useAutoDetect = False
        nic_spec.device.backing.network = network  # Sets port group to assign adapter to
        nic_spec.device.backing.deviceName = network.name  # Sets name of device on host system

        nic_spec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
        nic_spec.device.connectable.startConnected = True  # Ensures adapter is connected at boot
        nic_spec.device.connectable.allowGuestControl = True  # Allows guest OS to control device
        nic_spec.device.connectable.connected = True
        nic_spec.device.connectable.status = 'untried'
        # TODO: configure guest IP address if statically assigned
        self._edit(vim.vm.ConfigSpec(deviceChange=[nic_spec]))  # Apply change to VM

    def edit_nic(self, nic_id, port_group=None, summary=None):
        """
        Edits a VM NIC based on it's number
        :param nic_id: Number of network adapter on VM
        :param port_group: vim.Network object to assign NIC to [default: None]
        :param summary: Human-readable device description [default: None]
        """
        nic_label = 'Network adapter ' + str(nic_id)
        logging.debug("Changing '%s' on VM '%s'", nic_label, self.name)
        virtual_nic_device = self.get_nic_by_name(nic_label)
        if not virtual_nic_device:
            logging.error('Virtual %s could not be found!', nic_label)
            return
        nic_spec = vim.vm.device.VirtualDeviceSpec()
        nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.edit
        nic_spec.device = virtual_nic_device
        if summary:
            nic_spec.device.deviceInfo.summary = str(summary)
        if port_group:
            logging.debug("Changing PortGroup to: '%s'", port_group.name)
            nic_spec.device.backing.network = port_group
            nic_spec.device.backing.deviceName = port_group.name
        self._edit(vim.vm.ConfigSpec(deviceChange=[nic_spec]))  # Apply change to VM

    def delete_nic(self, nic_number):
        """
        Deletes VM vNIC based on it's number
        :param nic_number: Integer unit of the vNIC to delete
        """
        nic_label = 'Network adapter ' + str(nic_number)
        logging.debug("Removing Virtual %s from '%s'", nic_label, self.name)
        virtual_nic_device = self.get_nic_by_name(nic_label)
        if virtual_nic_device is not None:
            virtual_nic_spec = vim.vm.device.VirtualDeviceSpec()
            virtual_nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
            virtual_nic_spec.device = virtual_nic_device
            self._edit(vim.vm.ConfigSpec(deviceChange=[virtual_nic_spec]))  # Apply change to VM
        else:
            logging.error("Virtual %s could not be found for '%s'", nic_label, self.name)

    def attach_iso(self, iso_name, datastore, boot=True):
        """
        Attaches an ISO image to a VM
        :param iso_name: Name of the ISO image to attach
        :param datastore: vim.Datastore where the ISO resides
        :param boot: Set VM to boot from the attached ISO
        """
        logging.debug("Adding ISO '%s' to '%s'", iso_name, self.name)
        drive_spec = vim.vm.device.VirtualDeviceSpec()
        drive_spec.device = vim.vm.device.VirtualCdrom()
        drive_spec.device.key = -1
        drive_spec.device.unitNumber = 0

        # Find a disk controller to attach to
        controller = self._find_free_ide_controller()
        if controller:
            drive_spec.device.controllerKey = controller.key
        else:
            logging.error("Could not find a free IDE controller on '%s' to attach ISO '%s'",
                          self.name, iso_name)
            return

        drive_spec.device.backing = vim.vm.device.VirtualCdrom.IsoBackingInfo()
        drive_spec.device.backing.fileName = "[%s] %s" % (datastore.name, iso_name)  # Attach ISO
        drive_spec.device.backing.datastore = datastore  # Set datastore ISO is in

        drive_spec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
        drive_spec.device.connectable.allowGuestControl = True  # Allows guest OS to control device
        drive_spec.device.connectable.startConnected = True  # Ensures ISO is connected at boot

        drive_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
        vm_spec = vim.vm.ConfigSpec(deviceChange=[drive_spec])
        if boot:  # Set the VM to boot from the ISO upon power on
            logging.debug("Setting '%s' to boot from ISO '%s'", self.name, iso_name)
            order = [vim.vm.BootOptions.BootableCdromDevice()]
            order.extend(list(self._vm.config.bootOptions.bootOrder))
            vm_spec.bootOptions = vim.vm.BootOptions(bootOrder=order)
        self._edit(vm_spec)  # Apply change to VM

    def _find_free_ide_controller(self):
        """
        Finds a free IDE controller to use
        :return: vim.vm.device.VirtualIDEController
        """
        for dev in self._vm.config.hardware.device:
            if isinstance(dev, vim.vm.device.VirtualIDEController) and len(dev.device) < 2:
                return dev  # If there are less than 2 devices attached, we can use it
        return None

    def has_tools(self):
        """
        Checks if VMware Tools is installed and working
        :return: If tools are installed and working
        """
        tools = self._vm.summary.guest.toolsStatus
        return True if tools == "toolsOK" or tools == "toolsOld" else False

    def powered_on(self):
        """
        Determines if a VM is powered on
        :return: If VM is powered on
        """
        return self._vm.runtime.powerState == vim.VirtualMachine.PowerState.poweredOn

    def is_template(self):
        """
        Checks if VM is a template
        :return: If the VM is a template
        """
        return bool(self._vm.summary.config.template)

    def is_windows(self):
        """
        Checks if a VM's guest OS is Windows
        :return: Bool
        """
        return bool(str(self._vm.config.guestId).lower().startswith("win"))

    def _edit(self, config):
        try:
            wait_for_task(self._vm.ReconfigVM_Task(config))
        except vim.fault.TaskInProgress as e:
            logging.error("Cannot edit VM '%s': it is busy with task '%s'", self.name, e.task)
        except vim.fault.VmConfigFault:
            logging.error("Cannot edit VM '%s': invalid configuration", self.name)
        except vim.fault.InvalidState:
            logging.error("Cannot edit VM '%s': it is in an invalid state", self.name)
        except vim.fault.InvalidName as e:
            logging.error("Cannot edit VM '%s': name '%s' is not valid", self.name, e.name)
        except vim.fault.DuplicateName as e:
            logging.error("Cannot edit VM '%s': there is a duplicate with name '%s'", self.name,
                          e.name)
        except vim.fault.InvalidPowerState as e:
            logging.error("Cannot edit VM '%s': invalid power state '%s'", self.name,
                          e.existingState)
        except vim.fault.InvalidDatastore as e:
            logging.error("Cannot edit VM '%s': invalid Datastore '%s'", self.name, e.datastore)

    def __str__(self):
        return str(self.name)

    def __hash__(self):
        return hash(self._vm.summary.config.instanceUuid)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.name == other.name \
               and hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from pyVmomi import vim, vmodl

import adles.utils as utils
import adles.vsphere.vsphere_utils as vutils


@utils.check(vim.VirtualMachine, "vm")
@utils.time_execution
def clone_vm(vm, folder, name, clone_spec):
    """
    Creates a clone of the VM or Template
    :param vm: vim.VirtualMachine object
    :param folder: vim.Folder object in which to create the clone
    :param name: String name of the new VM
    :param clone_spec: vim.vm.CloneSpec for the new VM
    """
    if is_template(vm) and not bool(clone_spec.location.pool):
        logging.error("Cannot clone Template '%s' without specifying a resource pool", vm.name)
    else:
        logging.info("Cloning VM '%s' to folder '%s' as '%s'", vm.name, folder.name, name)
        try:
            vutils.wait_for_task(vm.CloneVM_Task(folder=folder, name=name, spec=clone_spec))
        except vim.fault.InvalidState:
            logging.error("Could not make clone '%s': invalid state for VM '%s'", name, vm.name)
        except vim.fault.CustomizationFault:
            logging.error("Could not make clone '%s': invalid customization", name)
        except vim.fault.VmConfigFault:
            logging.error("Could not make clone '%s': invalid configuration", name)


@utils.check(vim.Folder, "folder")
@utils.time_execution
def create_vm(folder, config, pool, host=None):
    """
    Creates a VM with the specified configuration in the given folder
    :param folder: vim.Folder in which to create the VM
    :param config: vim.vm.ConfigSpec defining the new VM
    :param pool: vim.ResourcePool to which the virtual machine will be attached
    :param host: vim.HostSystem on which the VM will run [default: None]
    """
    logging.info("Creating VM '%s' in folder '%s'", config.name, folder.name)
    vutils.wait_for_task(folder.CreateVM_Task(config, pool, host))


def gen_vm_spec(name, datastore_name, annotation=None, cpus=1, cores=1, memory=512,
                version=None, firmware='efi', max_consoles=None, datastore_path=None):
    """
    Generates a Virtual Machine creation spec
    :param name: Name of the VM
    :param datastore_name: Name of datastore to put VM on
    :param version: Hardware version of the VM [default: highest host supports]
    :param annotation: String annotation for the VM [default: None]
    :param cpus: Number of CPUs [default: 1]
    :param cores: Number of cores per CPU [default: 1]
    :param memory: Memory in MB [default: 512]
    :param firmware: (efi | bios) [default: efi]
    :param max_consoles: Maximum number of active remote display connections
    [default: datacenter-defined]
    :param datastore_path: Path to existing VM files on datastore [default: None]
    :return: vim.vm.ConfigSpec
    """
    spec = vim.vm.ConfigSpec()
    spec.name = str(name)
    spec.numCPUs = int(cpus)
    spec.numCoresPerSocket = int(cores)
    spec.memoryMB = int(memory)
    spec.memoryHotAddEnabled = True
    spec.firmware = str(firmware).lower()

    if version:
        spec.version = str(version)

    if annotation:
        spec.annotation = str(annotation)

    if max_consoles:
        spec.maxMksConnections = int(max_consoles)

    vm_path = '[' + datastore_name + '] '
    if datastore_path:
        vm_path += str(datastore_path)
    vm_path += name + '/' + name + '.vmx'
    spec.files = vim.vm.FileInfo(vmPathName=vm_path)
    return spec


@utils.check(vim.VirtualMachine, "vm")
@utils.time_execution
def destroy_vm(vm):
    """
    Destroys a Virtual Machine
    :param vm: vim.VirtualMachine object
    """
    name = vm.name
    logging.info("Destroying VM '%s'", name)
    if powered_on(vm):
        logging.info("VM '%s' is still on, powering off before destroying...", name)
        change_vm_state(vm, "off", attempt_guest=False)
    vutils.wait_for_task(vm.Destroy_Task())


@utils.check(vim.VirtualMachine, "vm")
def edit_vm(vm, config):
    """
    Edits a VM using the given configuration specification
    :param vm: vim.VirtualMachine object
    :param config: vim.vm.ConfigSpec object
    """
    logging.debug("Reconfiguring VM '%s'", vm.name)
    try:
        vutils.wait_for_task(vm.ReconfigVM_Task(config))
    except vim.fault.TaskInProgress as e:
        logging.error("Cannot edit VM '%s': it is busy with task '%s'", vm.name, e.task)
    except vim.fault.VmConfigFault:
        logging.error("Cannot edit VM '%s': invalid configuration", vm.name)
    except vim.fault.InvalidState:
        logging.error("Cannot edit VM '%s': it is in an invalid state", vm.name)
    except vim.fault.InvalidName as e:
        logging.error("Cannot edit VM '%s': name '%s' is not valid", vm.name, e.name)
    except vim.fault.DuplicateName as e:
        logging.error("Cannot edit VM '%s': there is a duplicate with name '%s'", vm.name, e.name)
    except vim.fault.InvalidPowerState as e:
        logging.error("Cannot edit VM '%s': invalid power state '%s'", vm.name, e.existingState)
    except vim.fault.InvalidDatastore as e:
        logging.error("Cannot edit VM '%s': invalid Datastore '%s'", vm.name, e.datastore)


@utils.check(vim.VirtualMachine, "vm")
def upgrade_vm(vm, version=''):
    """
    Upgrades the hardware version of the VM
    :param vm: vim.VirtualMachine object
    :param version: Version of hardware to upgrade VM to [default: latest VM's host supports]
    """
    try:
        vutils.wait_for_task(vm.UpgradeVM(str(version)))
    except vim.fault.AlreadyUpgraded:
        logging.error("VM '%s' hardware version is already up-to-date", vm.name)
    except vim.fault.InvalidPowerState as e:
        logging.error("Cannot upgrade VM '%s': invalid power state '%s'", vm.name, e.existingState)


@utils.check(vim.VirtualMachine, "vm")
def change_vm_state(vm, state, attempt_guest=True):
    """
    :param vm: vim.VirtualMachine object to change state of
    :param state: State to change to (on | off | reset | suspend)
    :param attempt_guest: Whether to attempt to use guest operations to change power state
    """
    if is_template(vm):
        logging.error("VM '%s' is a Template, so state cannot be changed to '%s'", vm.name, state)
        return
    try:
        if attempt_guest and has_tools(vm) and state != "on":  # Can't power on using guest ops
            change_guest_state(vm, state)
        else:
            change_power_state(vm, state)
    except vim.fault.InvalidPowerState as e:
        logging.error("Cannot change '%s' in power state '%s' to '%s'",
                      vm.name, e.existingState, state)
    except vmodl.fault.NotSupported:
        logging.error("Cannot change power state of '%s': it is a template", vm.name)
    except vim.fault.TaskInProgress as e:
        logging.error("Cannot change power state of '%s': Task '%s' in progress",
                      vm.name, e.task)
    except vim.fault.InvalidState:
        logging.error("Cannot change power state of '%s': it is in an invalid state ", vm.name)


@utils.check(vim.VirtualMachine, "vm")
def change_power_state(vm, power_state):
    """
    Changes a VM power state to the state specified
    :param vm: vim.VirtualMachine object to change power state of
    :param power_state: on, off, reset, or suspend
    """
    state = power_state.lower()
    if state == "on":
        task = vm.PowerOnVM_Task()
    elif state == "off":
        task = vm.PowerOffVM_Task()
    elif state == "reset":
        task = vm.ResetVM_Task()
    elif state == "suspend":
        task = vm.SuspendVM_Task()
    else:
        logging.error("Invalid power_state %s for VM %s", power_state, vm.name)
        return
    logging.debug("Changing power state of VM %s to: '%s'", vm.name, power_state)
    vutils.wait_for_task(task)


@utils.check(vim.VirtualMachine, "vm")
def change_guest_state(vm, guest_state):
    """
    Changes a VMs guest power state. VMware Tools must be installed on the VM for this to work.
    :param vm:  vim.VirtualMachine object to change guest state of
    :param guest_state: shutdown, reboot, or standby [Alternatives: off, reset, suspend]
    """
    state = guest_state.lower()
    if vm.summary.guest.toolsStatus == "toolsNotInstalled":
        logging.error("Cannot change a VM's guest power state without VMware Tools!")
        return
    elif state == "shutdown" or state == "off":
        task = vm.ShutdownGuest()
    elif state == "reboot" or state == "reset":
        task = vm.RebootGuest()
    elif state == "standby" or state == "suspend":
        task = vm.StandbyGuest()
    else:
        logging.error("Invalid guest_state argument: %s", state)
        return
    logging.debug("Changing guest power state of VM %s to: '%s'", vm.name, guest_state)
    try:
        vutils.wait_for_task(task)
    except vim.fault.ToolsUnavailable:
        logging.error("Cannot change guest state of '%s': Tools are not running", vm.name)


@utils.check(vim.VirtualMachine, "vm")
def convert_to_template(vm):
    """
    Converts a Virtual Machine to a template
    :param vm: vim.VirtualMachine object to convert
    """
    if is_template(vm):
        logging.warning("VM '%s' is already a template", vm.name)
        return
    try:
        logging.debug("Converting VM '%s' to Template", vm.name)
        vm.MarkAsTemplate()
    except vim.fault.InvalidPowerState as e:
        logging.error("Cannot convert '%s' to a template while in state '%s'",
                      vm.name, e.existingState)
    except vim.fault.InvalidState:
        logging.error("Cannot '%s' to template: it is in an invalid state", vm.name)


@utils.check(vim.VirtualMachine, "vm")
def convert_to_vm(vm, resource_pool, host=None):
    """
    Converts a Template to a Virtual Machine
    :param vm: vim.VirtualMachine object to convert
    :param resource_pool: vim.ResourcePool to associate with the VM
    :param host: vim.HostSystem on which the VM should run [default: None]
    """
    logging.debug("Converting Template '%s' to VM and assigning to resource pool '%s'",
                  vm.name, resource_pool.name)
    vm.MarkAsVirtualMachine(resource_pool, host)


@utils.check(vim.VirtualMachine, "vm")
def set_note(vm, note):
    """
    Sets the note on the VM to note
    :param vm: vim.VirtualMachine object
    :param note: String to set the note to
    """
    logging.debug("Setting note of VM '%s' to '%s'", vm.name, note)
    spec = vim.vm.ConfigSpec()
    spec.annotation = note
    edit_vm(vm, spec)


@utils.check(vim.VirtualMachine, "vm")
@utils.time_execution
def create_snapshot(vm, name, description='default', memory=False):
    """
    Create a snapshot of the VM
    :param vm: vim.VirtualMachine object
    :param name: Title of the snapshot
    :param memory: Memory dump of the VM is included in the snapshot
    :param description: Text description of the snapshot
    """
    logging.info("Creating snapshot '%s' of '%s'", name, vm.name)
    vutils.wait_for_task(vm.CreateSnapshot_Task(name=name, description=description,
                                                memory=bool(memory), quiesce=True))


@utils.check(vim.VirtualMachine, "vm")
@utils.time_execution
def revert_to_snapshot(vm, snapshot_name):
    """
    Reverts VM to the named snapshot
    :param vm: vim.VirtualMachine object
    :param snapshot_name: Name of the snapshot to revert to
    """
    logging.info("Reverting '%s' to the snapshot '%s'", vm.name, snapshot_name)
    snap = get_snapshot(vm, snapshot_name)
    vutils.wait_for_task(snap.RevertToSnapshot_Task())


@utils.check(vim.VirtualMachine, "vm")
@utils.time_execution
def revert_to_current_snapshot(vm):
    """
    Reverts the VM to the most recent snapshot
    :param vm: vim.VirtualMachine object
    """
    logging.info("Reverting '%s' to the current snapshot", vm.name)
    vutils.wait_for_task(vm.RevertToCurrentSnapshot_Task())


@utils.check(vim.VirtualMachine, "vm")
def get_snapshot(vm, snapshot_name):
    """
    Retrieves the named snapshot from the VM
    :param vm: vim.VirtualMachine object
    :param snapshot_name: Name of the snapshot to get
    :return: vim.Snapshot object
    """
    for snap in get_all_snapshots(vm):
        if snap.name == snapshot_name:
            return snap.snapshot
    return None


@utils.check(vim.VirtualMachine, "vm")
def get_current_snapshot(vm):
    """
    Retrieves the current snapshot from the VM
    :param vm: vim.VirtualMachine object
    :return: Current vim.Snapshot object for the VM, or None if there are no snapshots
    """
    return vm.snapshot.currentSnapshot


@utils.check(vim.VirtualMachine, "vm")
def get_all_snapshots(vm):
    """
    Retrieves a list of all snapshots of the VM
    :param vm: vim.VirtualMachine object
    :return: Nested List of vim.Snapshot objects
    """
    return _get_snapshots_recursive(vm.snapshot.rootSnapshotList)


# From: https://github.com/imsweb/ezmomi
def _get_snapshots_recursive(snap_tree):
    """
    Recursively finds all snapshots in the tree
    :param snap_tree: Tree of snapshots
    :return: Nested List of vim.Snapshot objects
    """
    local_snap = []
    for snap in snap_tree:
        local_snap.append(snap)
    for snap in snap_tree:
        recurse_snap = _get_snapshots_recursive(snap.childSnapshotList)
        if recurse_snap:
            local_snap.extend(recurse_snap)
    return local_snap


@utils.check(vim.VirtualMachine, "vm")
def remove_snapshot(vm, snapshot_name, remove_children=True, consolidate_disks=True):
    """
    Removes the named snapshot from the VM
    :param vm: vim.VirtualMachine object
    :param snapshot_name: Name of the snapshot to remove
    :param remove_children: Removal of the entire snapshot subtree [default: True]
    :param consolidate_disks: Virtual disks of deleted snapshot will be merged with
    other disks if possible [default: True]
    """
    logging.info("Removing snapshot '%s' from '%s'", snapshot_name, vm.name)
    snapshot = get_snapshot(vm, snapshot_name)
    vutils.wait_for_task(snapshot.RemoveSnapshot_Task(remove_children, consolidate_disks))


@utils.check(vim.VirtualMachine, "vm")
def remove_all_snapshots(vm, consolidate_disks=True):
    """
    Removes all snapshots associated with the VM
    :param vm: vim.VirtualMachine object
    :param consolidate_disks: Virtual disks of the deleted snapshot will be merged with
    other disks if possible [default: True]
    """
    logging.info("Removing ALL snapshots for the '%s'", vm.name)
    vutils.wait_for_task(vm.RemoveAllSnapshots_Task(consolidate_disks))


@utils.check(vim.VirtualMachine, "vm")
def snapshot_disk_usage(vm):
    """
    Determines the total disk usage of a VM's snapshots
    :param vm: vim.VirtualMachine
    :return: Human-readable disk usage of the snapshots
    """
    from re import search
    disk_list = vm.layoutEx.file
    size = 0
    for disk in disk_list:
        if disk.type == 'snapshotData':
            size += disk.size
        ss_disk = search('0000\d\d', disk.name)
        if ss_disk:
            size += disk.size
    return utils.sizeof_fmt(size)


# From: getallvms in pyvmomi-community-samples
@utils.check(vim.VirtualMachine, "vm")
def get_vm_info(vm, detailed=False, uuids=False, snapshot=False, vnics=False):
    """
    Get human-readable information for a VM
    :param vm: vim.VirtualMachine object
    :param detailed: Add more detailed information, such as maximum memory used [default: False]
    :param uuids: Whether to get UUID information [default: False]
    :param snapshot: Shows the current snapshot, if any [default: False]
    :param vnics: Add information about the virtual network interfaces on the VM
    :return: String with the VM information
    """
    info_string = "\n"
    summary = vm.summary
    info_string += "Name          : %s\n" % summary.config.name
    info_string += "Status        : %s\n" % str(summary.overallStatus)
    info_string += "Power State   : %s\n" % summary.runtime.powerState
    if vm.guest:
        info_string += "Guest State   : %s\n" % vm.guest.guestState
    info_string += "Last modified : %s\n" % str(vm.config.modified)  # datetime object
    if detailed:
        info_string += "Num consoles  : %d\n" % summary.runtime.numMksConnections
    info_string += "Host          : %s\n" % summary.runtime.host.name
    info_string += "Guest OS      : %s\n" % summary.config.guestFullName
    info_string += "Num CPUs      : %s\n" % summary.config.numCpu
    info_string += "Memory (MB)   : %s\n" % summary.config.memorySizeMB
    if detailed:
        info_string += "Num vNICs     : %s\n" % summary.config.numEthernetCards
        info_string += "Num Disks     : %s\n" % summary.config.numVirtualDisks
    info_string += "IsTemplate    : %s\n" % summary.config.template  # bool
    if detailed:
        info_string += "Path          : %s\n" % summary.config.vmPathName
    info_string += "Folder:       : %s\n" % vm.parent.name
    if vm.guest:
        info_string += "IP            : %s\n" % vm.guest.ipAddress
        info_string += "Hostname:     : %s\n" % vm.guest.hostName
        info_string += "Tools status  : %s\n" % vm.guest.toolsRunningStatus
        info_string += "Tools version : %s\n" % vm.guest.toolsVersionStatus2
    if vnics:
        vm_nics = get_nics(vm)
        for num, vnic in zip(range(1, len(vm_nics) + 1), vm_nics):
            info_string += "vNIC %d label   : %s\n" % (num, vnic.deviceInfo.label)
            info_string += "vNIC %d summary : %s\n" % (num, vnic.deviceInfo.summary)
            info_string += "vNIC %d network : %s\n" % (num, vnic.backing.network.name)
    if uuids:
        info_string += "Instance UUID : %s\n" % summary.config.instanceUuid
        info_string += "Bios UUID     : %s\n" % summary.config.uuid
    if summary.runtime.question:
        info_string += "Question      : %s\n" % summary.runtime.question.text
    if summary.config.annotation:
        info_string += "Annotation    : %s\n" % summary.config.annotation
    if snapshot and vm.snapshot and hasattr(vm.snapshot, 'currentSnapshot'):
        info_string += "Current Snapshot: %s\n" % vm.snapshot.currentSnapshot.config.name
        info_string += "Disk usage of all snapshots: %s\n" % snapshot_disk_usage(vm=vm)
    if detailed and summary.runtime:
        info_string += "Last Poweron  : %s\n" % str(summary.runtime.bootTime)  # datetime object
        info_string += "Max CPU usage : %s\n" % summary.runtime.maxCpuUsage
        info_string += "Max Mem usage : %s\n" % summary.runtime.maxMemoryUsage
        info_string += "Last suspended: %s\n" % summary.runtime.suspendTime
    return info_string


@utils.check(vim.VirtualMachine, "vm")
def remove_device(vm, device):
    """
    Removes the device from the VM
    :param vm: vim.VirtualMachine
    :param device: vim.vm.device.VirtualDeviceSpec
    """
    logging.debug("Removing device '%s' from VM '%s'", device.name, vm.name)
    device.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
    edit_vm(vm, vim.vm.ConfigSpec(deviceChange=[device]))


@utils.check(vim.VirtualMachine, "vm")
def get_nics(vm):
    """
    Returns a list of all Virtual Network Interface Cards (vNICs) on a VM
    :param vm: vim.VirtualMachine
    :return: list of vim.vm.device.VirtualEthernetCard
    """
    return [dev for dev in vm.config.hardware.device if vutils.is_vnic(dev)]


@utils.check(vim.VirtualMachine, "vm")
def get_nic_by_name(vm, name):
    """
    Gets a Virtual Network Interface Card from a VM
    :param vm: vim.VirtualMachine
    :param name: Name of the vNIC
    :return: vim.vm.device.VirtualEthernetCard
    """
    for dev in vm.config.hardware.device:
        if vutils.is_vnic(dev) and dev.deviceInfo.label.lower() == name.lower():
            return dev
    logging.debug("Could not find vNIC '%s' on VM '%s'", name, vm.name)
    return None


@utils.check(vim.VirtualMachine, "vm")
def get_nic_by_id(vm, nic_id):
    """
    Get a vNIC by integer ID
    :param vm: vim.VirtualMachine
    :param nic_id: ID of the vNIC
    :return: vim.vm.device.VirtualEthernetCard
    """
    return get_nic_by_name(vm, "Network Adapter " + str(nic_id))


@utils.check(vim.VirtualMachine, "vm")
def get_nic_by_network(vm, network):
    """
    Finds a vNIC by it's network backing
    :param vm: vim.VirtualMachine
    :param network: vim.Network
    :return: Name of the vNIC
    """
    for dev in vm.config.hardware.device:
        if vutils.is_vnic(dev) and dev.backing.network == network:
            return dev
    logging.debug("Could not find vNIC with network '%s' on VM '%s'", network.name, vm.name)
    return None


# From: add_nic_to_vm in pyvmomi-community-samples
@utils.check(vim.VirtualMachine, "vm")
def add_nic(vm, network, summary="default-summary", model="e1000"):
    """
    Add a NIC in the portgroup to the VM
    :param vm: vim.VirtualMachine
    :param network: vim.Network to attach NIC to
    :param summary: Human-readable device info [default: default-summary]
    :param model: Model of virtual network adapter. [default: e1000]
    Options: (e1000 | e1000e | vmxnet | vmxnet2 | vmxnet3)
    e1000 will work on Windows Server 2003+, and e1000e is supported on Windows Server 2012+.
    VMXNET adapters require VMware Tools to be installed, and provide enhanced performance.
    Read this for more details: http://rickardnobel.se/vmxnet3-vs-e1000e-and-e1000-part-1/
    """
    if not isinstance(network, vim.Network):
        logging.error("Invalid network type when adding vNIC to VM '%s': %s",
                      vm.name, type(network).__name__)
    logging.debug("Adding NIC to VM '%s'\nNetwork: '%s'\tSummary: '%s'\tNIC Model: '%s'",
                  vm.name, network.name, summary, model)
    nic_spec = vim.vm.device.VirtualDeviceSpec()  # Create a base object to add configurations to
    nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add

    # Set the type of network adapter
    if model == "e1000":
        nic_spec.device = vim.vm.device.VirtualE1000()
    elif model == "e1000e":
        nic_spec.device = vim.vm.device.VirtualE1000e()
    elif model == "vmxnet":
        nic_spec.device = vim.vm.device.VirtualVmxnet()
    elif model == "vmxnet2":
        nic_spec.device = vim.vm.device.VirtualVmxnet2()
    elif model == "vmxnet3":
        nic_spec.device = vim.vm.device.VirtualVmxnet3()
    else:
        logging.error("Invalid NIC model: '%s'\nDefaulting to e1000...", model)
        nic_spec.device = vim.vm.device.VirtualE1000()
    nic_spec.device.addressType = 'generated'  # Sets how MAC address is assigned
    nic_spec.device.wakeOnLanEnabled = False  # Disables Wake-on-lan capabilities

    nic_spec.device.deviceInfo = vim.Description()
    nic_spec.device.deviceInfo.summary = summary

    nic_spec.device.backing = vim.vm.device.VirtualEthernetCard.NetworkBackingInfo()
    nic_spec.device.backing.useAutoDetect = False
    nic_spec.device.backing.network = network  # Sets port group to assign adapter to
    nic_spec.device.backing.deviceName = network.name  # Sets name of device on host system

    nic_spec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
    nic_spec.device.connectable.startConnected = True  # Ensures adapter is connected at boot
    nic_spec.device.connectable.allowGuestControl = True  # Allows guest OS to control device
    nic_spec.device.connectable.connected = True
    nic_spec.device.connectable.status = 'untried'
    # TODO: configure guest IP address if statically assigned

    edit_vm(vm, vim.vm.ConfigSpec(deviceChange=[nic_spec]))  # Apply the change to the VM


@utils.check(vim.VirtualMachine, "vm")
def edit_nic(vm, nic_id, port_group=None, summary=None):
    """
    Edits a VM NIC based on it's number
    :param vm: vim.VirtualMachine
    :param nic_id: Number of network adapter on VM
    :param port_group: vim.Network object to assign NIC to [default: None]
    :param summary: Human-readable device description [default: None]
    """
    nic_label = 'Network adapter ' + str(nic_id)
    logging.debug("Changing '%s' on VM '%s'", nic_label, vm.name)
    virtual_nic_device = get_nic_by_name(vm, nic_label)
    if not virtual_nic_device:
        logging.error('Virtual %s could not be found!', nic_label)
        return

    nic_spec = vim.vm.device.VirtualDeviceSpec()
    nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.edit
    nic_spec.device = virtual_nic_device
    if summary:
        nic_spec.device.deviceInfo.summary = str(summary)
    if port_group:
        logging.debug("Changing PortGroup to: '%s'", port_group.name)
        nic_spec.device.backing.network = port_group
        nic_spec.device.backing.deviceName = port_group.name
    edit_vm(vm, vim.vm.ConfigSpec(deviceChange=[nic_spec]))  # Apply the change to the VM


# From: delete_nic_from_vm in pyvmomi-community-samples
@utils.check(vim.VirtualMachine, "vm")
def delete_nic(vm, nic_number):
    """
    Deletes VM vNIC based on it's number
    :param vm: vim.VirtualMachine
    :param nic_number: Integer unit of the vNIC to delete
    """
    nic_label = 'Network adapter ' + str(nic_number)
    logging.debug("Removing Virtual %s from '%s'", nic_label, vm.name)
    virtual_nic_device = get_nic_by_name(vm, nic_label)

    if not virtual_nic_device:
        logging.error("Virtual %s could not be found for VM '%s'", nic_label, vm.name)
        return

    virtual_nic_spec = vim.vm.device.VirtualDeviceSpec()
    virtual_nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
    virtual_nic_spec.device = virtual_nic_device

    edit_vm(vm, vim.vm.ConfigSpec(deviceChange=[virtual_nic_spec]))  # Apply the change to the VM


@utils.check(vim.VirtualMachine, "vm")
def attach_iso(vm, iso_name, datastore, boot=True):
    """
    Attaches an ISO image to a VM
    :param vm: vim.VirtualMachine
    :param iso_name: Name of the ISO image to attach
    :param datastore: vim.Datastore where the ISO resides
    :param boot: Set VM to boot from the attached ISO
    """
    logging.debug("Adding ISO '%s' to VM '%s'", iso_name, vm.name)
    drive_spec = vim.vm.device.VirtualDeviceSpec()
    drive_spec.device = vim.vm.device.VirtualCdrom()
    drive_spec.device.key = -1
    drive_spec.device.unitNumber = 0

    # Find a disk controller to attach to
    controller = _find_free_ide_controller(vm)
    if controller:
        drive_spec.device.controllerKey = controller.key
    else:
        logging.error("Could not find a free IDE controller on VM '%s' to attach ISO '%s'",
                      vm.name, iso_name)
        return

    drive_spec.device.backing = vim.vm.device.VirtualCdrom.IsoBackingInfo()
    drive_spec.device.backing.fileName = "[%s] %s" % (datastore.name, iso_name)  # Attach ISO
    drive_spec.device.backing.datastore = datastore  # Set datastore ISO is in

    drive_spec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
    drive_spec.device.connectable.allowGuestControl = True  # Allows guest OS to control device
    drive_spec.device.connectable.startConnected = True     # Ensures ISO is connected at boot

    drive_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
    vm_spec = vim.vm.ConfigSpec(deviceChange=[drive_spec])

    if boot:  # Set the VM to boot from the ISO upon power on
        logging.debug("Setting '%s' to boot from ISO '%s'", vm.name, iso_name)
        order = [vim.vm.BootOptions.BootableCdromDevice()]
        order.extend(list(vm.config.bootOptions.bootOrder))
        vm_spec.bootOptions = vim.vm.BootOptions(bootOrder=order)

    edit_vm(vm, vm_spec)  # Apply the change to the VM


# From: cdrom_vm in pyvmomi-community-samples
@utils.check(vim.VirtualMachine, "vm")
def _find_free_ide_controller(vm):
    """
    Finds a free IDE controller to use
    :param vm: vim.VirtualMachine
    :return: vim.vm.device.VirtualIDEController
    """
    for dev in vm.config.hardware.device:
        if isinstance(dev, vim.vm.device.VirtualIDEController) and len(dev.device) < 2:
            return dev  # If there are less than 2 devices attached, we can use it
    return None


@utils.check(vim.VirtualMachine, "vm")
def has_tools(vm):
    """
    Checks if VMware Tools is installed and working
    :param vm: vim.VirtualMachine
    :return: If tools are installed and working
    """
    tools = vm.summary.guest.toolsStatus
    return True if tools == "toolsOK" or tools == "toolsOld" else False


@utils.check(vim.VirtualMachine, "vm")
def powered_on(vm):
    """
    Determines if a VM is powered on
    :param vm: vim.VirtualMachine object
    :return: If VM is powered on
    """
    return vm.runtime.powerState == vim.VirtualMachine.PowerState.poweredOn


@utils.check(vim.VirtualMachine, "vm")
def is_template(vm):
    """
    Checks if VM is a template
    :param vm: vim.VirtualMachine
    :return: If the VM is a template
    """
    return bool(vm.summary.config.template)


@utils.check(vim.VirtualMachine, "vm")
def is_windows(vm):
    """
    Checks if a VM's guest OS is Windows
    :param vm: vim.VirtualMachine
    :return: Bool
    """
    return bool(str(vm.config.guestId).lower().startswith("win"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Clone multiple Virtual Machines in vSphere.

Usage:
    clone-vms [options]

Options:
    -h, --help          Prints this page
    --version           Prints current version
    -n, --no-color      Do not color terminal output
    -v, --verbose       Emit debugging logs to terminal
    -f, --file FILE     Name of JSON file with server connection information

Examples:
    clone-vms -vf logins.json

"""

import logging

from docopt import docopt

from adles.utils import prompt_y_n_question, pad, default_prompt, script_setup, resolve_path
from adles.vsphere.vm_utils import clone_vm
from adles.vsphere.folder_utils import format_structure, retrieve_items

__version__ = "0.5.3"


def main():
    args = docopt(__doc__, version=__version__, help=True)
    server = script_setup('clone_vms.log', args, (__file__, __version__))

    vms = []
    vm_names = []

    # Single-vm source
    if prompt_y_n_question("Do you want to clone from a single VM?"):
        v, v_name = resolve_path(server, "VM", "or template you wish to clone")
        vms.append(v)
        vm_names.append(str(input("Base name for instances to be created: ")))
    # Multi-VM source
    else:
        folder_from, from_name = resolve_path(server, "folder", "you want to clone all VMs in")
        v, _ = retrieve_items(folder_from)  # Get VMs in the folder, ignore any folders
        vms.extend(v)
        logging.info("%d VMs found in source folder %s\n%s",
                     len(v), from_name, format_structure(v))
        if not prompt_y_n_question("Keep the same names? "):
            names = []
            for i in range(len(v)):
                names.append(str(input("Enter base name for VM %d: " % i)))
        else:
            names = list(map(lambda x: x.name, v))  # Same names as sources
        vm_names.extend(names)

    create_in, create_in_name = resolve_path(server, "folder", "in which to create VMs")
    instance_folder_base = None
    if prompt_y_n_question("Do you want to create a folder for each instance? "):
        instance_folder_base = str(input("Enter instance folder base name: "))

    num_instances = int(input("Number of instances to be created: "))

    pool = server.get_pool().name
    pool = default_prompt(prompt="Resource pool to assign VMs to", default=pool)

    logging.info("Creating %d instances under folder %s", num_instances, create_in_name)
    for instance in range(num_instances):
        for vm, name in zip(vms, vm_names):
            spec = server.gen_clone_spec(pool_name=pool)  # Generate clone specification
            if instance_folder_base:  # Create instance folders for a nested clone
                f = server.create_folder(instance_folder_base + pad(instance), create_in=create_in)
                vm_name = name
                clone_vm(vm=vm, folder=f, name=vm_name, clone_spec=spec)
            else:
                vm_name = name + pad(instance)    # Append instance number
                clone_vm(vm=vm, folder=create_in, name=vm_name, clone_spec=spec)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cleanup and Destroy VMs and VM Folders in a vSphere environment.

Usage:
    cleanup-vms [options]

Options:
    -h, --help          Prints this page
    --version           Prints current version
    -n, --no-color      Do not color terminal output
    -v, --verbose       Emit debugging logs to terminal
    -f, --file FILE     Name of JSON file with server connection information

Examples:
    cleanup-vms -vf logins.json

"""

import logging

from docopt import docopt

from adles.utils import prompt_y_n_question, default_prompt, script_setup, resolve_path
import adles.vsphere.vm_utils as vm_utils
from adles.vsphere.folder_utils import enumerate_folder, \
    format_structure, cleanup, retrieve_items

__version__ = "0.5.7"


def main():
    args = docopt(__doc__, version=__version__, help=True)
    server = script_setup('cleanup_vms.log', args, (__file__, __version__))

    if prompt_y_n_question("Multiple VMs? ", default="yes"):
        folder, folder_name = resolve_path(server, "folder",
                                           "that has the VMs/folders you want to destroy")

        # Display folder structure
        if prompt_y_n_question("Display the folder structure? "):
            logging.info("Folder structure: \n%s", format_structure(
                enumerate_folder(folder, recursive=True, power_status=True)))

        # Prompt user to configure destruction options
        print("Answer the following questions to configure the cleanup")
        if prompt_y_n_question("Destroy everything in and including the folder? "):
            vm_prefix = ''
            folder_prefix = ''
            recursive = True
            destroy_folders = True
            destroy_self = True
        else:
            vm_prefix = default_prompt("Prefix of VMs you wish to destroy"
                                       " (CASE SENSITIVE!)", default='')
            recursive = prompt_y_n_question("Recursively descend into folders? ")
            destroy_folders = prompt_y_n_question("Destroy folders in addition to VMs? ")
            if destroy_folders:
                folder_prefix = default_prompt("Prefix of folders you wish to destroy"
                                               " (CASE SENSITIVE!)", default='')
                destroy_self = prompt_y_n_question("Destroy the folder itself? ")
            else:
                folder_prefix = ''
                destroy_self = False

        # Show user what options they selected (TODO: do these options as cmdline arg flags)
        logging.info("Options selected\nVM Prefix: %s\nFolder Prefix: %s\nRecursive: %s\n"
                     "Folder-destruction: %s\nSelf-destruction: %s", str(vm_prefix),
                     str(folder_prefix), recursive, destroy_folders, destroy_self)

        # Show how many items matched the options
        v, f = retrieve_items(folder, vm_prefix, folder_prefix, recursive=True)
        num_vms = len(v)
        if destroy_folders:
            num_folders = len(f)
            if destroy_self:
                num_folders += 1
        else:
            num_folders = 0
        logging.info("%d VMs and %d folders match the options", num_vms, num_folders)

        # Confirm and destroy
        if prompt_y_n_question("Continue with destruction? "):
            logging.info("Destroying folder '%s'...", folder_name)
            cleanup(folder, vm_prefix=vm_prefix, folder_prefix=folder_prefix, recursive=recursive,
                    destroy_folders=destroy_folders, destroy_self=destroy_self)
        else:
            logging.info("Destruction cancelled")
    else:
        vm, vm_name = resolve_path(server, "vm", "to destroy")

        if prompt_y_n_question("Display VM info? "):
            logging.info(vm_utils.get_vm_info(vm, detailed=True, uuids=True,
                                              snapshot=True, vnics=True))

        if vm.config.template:  # Warn if template
            if not prompt_y_n_question("VM '%s' is a Template. Continue? " % vm_name):
                exit(0)

        if prompt_y_n_question("Continue with destruction? "):
            logging.info("Destroying VM '%s'", vm_name)
            vm_utils.destroy_vm(vm)
        else:
            logging.info("Destruction cancelled")

if __name__ == '__main__':
    main()
